import paynt.quotient
import paynt.quotient.fsc
import stormpy
import stormpy.pomdp
import payntbind

import paynt.utils.timer

from os import makedirs

from threading import Thread
from time import sleep

import logging
logger = logging.getLogger(__name__)


# class implementing the main components of the Storm integration for FSC synthesis for POMDPs
class StormPOMDPControl:

    def __init__(self):
        self.latest_storm_result = None      # holds object representing the latest Storm result
        self.storm_bounds = None             # under-approximation value from Storm

        self.saynt_fsc = None                # holds the FSC synthesized using SAYNT

        # PAYNT data and FSC export
        self.latest_paynt_result = None      # holds the synthesised assignment
        self.latest_paynt_result_fsc = None  # holds the FSC built from assignment
        self.paynt_bounds = None
        self.paynt_export = []

        # parsed best result data dictionary (Starting with data from Storm)
        self.result_dict = {}
        self.result_dict_no_cutoffs = {}
        self.result_dict_paynt = {}
        self.memory_vector = {}

        # controller sizes
        self.belief_controller_size = None
        self.paynt_fsc_size = None

        self.is_storm_better = False

        self.pomdp = None                    # The original POMDP model
        self.quotient = None
        self.spec_formulas = None            # The specification to be checked
        self.storm_options = None
        self.get_result = None
        self.use_cutoffs = False
        self.unfold_strategy_storm = None

        # PAYNT/Storm iteration settings
        self.iteration_timeout = None
        self.paynt_timeout = None
        self.storm_timeout = None

        self.storm_terminated = False

        self.s_queue = None

        self.saynt_timer = None

    def set_options(self,
        storm_options, get_storm_result, iterative_storm, use_storm_cutoffs,
        unfold_strategy_storm, prune_storm
    ):
        self.storm_options = storm_options
        if get_storm_result is not None:
            self.get_result = get_storm_result
        if iterative_storm is not None:
            self.iteration_timeout, self.paynt_timeout, self.storm_timeout = iterative_storm
        self.use_cutoffs = use_storm_cutoffs
        self.unfold_strategy_storm = unfold_strategy_storm

        self.incomplete_exploration = False
        if prune_storm:
            self.incomplete_exploration = True

        self.unfold_storm = True
        self.unfold_cutoff = False
        if unfold_strategy_storm == "paynt":
            self.unfold_storm = False
        elif unfold_strategy_storm == "cutoff":
            self.unfold_cutoff = True

    def get_storm_result(self):
        self.run_storm_analysis()
        self.parse_results(self.quotient)
        self.update_data()

        if self.s_queue is not None:
            self.s_queue.put((self.result_dict, self.storm_bounds))

    def store_storm_result(self, result):
        self.latest_storm_result = result
        if self.quotient.specification.optimality.minimizing:
            self.storm_bounds = self.latest_storm_result.upper_bound
        else:
            self.storm_bounds = self.latest_storm_result.lower_bound

        if (self.iteration_timeout is not None) or (self.get_result is not None):
            self.saynt_fsc = self.belief_controller_to_fsc(self.latest_storm_result, self.latest_paynt_result_fsc)

    # run Storm POMDP analysis for given model and specification
    # TODO: discuss Storm options
    def run_storm_analysis(self):
        # TODO rework options
        if self.storm_options == "cutoff":
            options = self.get_cutoff_options(100000)
        elif self.storm_options == "clip2":
            options = self.get_clip2_options()
        elif self.storm_options == "clip4":
            options = self.get_clip4_options()
        elif self.storm_options == "small":
            options = self.get_cutoff_options(100)
        elif self.storm_options == "2mil":
            options = self.get_cutoff_options(2000000)
        elif self.storm_options == "5mil":
            options = self.get_cutoff_options(5000000)
        elif self.storm_options == "10mil":
            options = self.get_cutoff_options(10000000)
        elif self.storm_options == "20mil":
            options = self.get_cutoff_options(20000000)
        elif self.storm_options == "30mil":
            options = self.get_cutoff_options(30000000)
        elif self.storm_options == "50mil":
            options = self.get_cutoff_options(50000000)
        elif self.storm_options == "refine":
            options = self.get_refine_options()
        elif self.storm_options == "overapp":
            options = self.get_overapp_options()
        else:
            logger.error("Incorrect option setting for Storm was found")
            exit()

        belmc = stormpy.pomdp.BeliefExplorationModelCheckerDouble(self.pomdp, options)

        logger.info("starting Storm POMDP analysis")
        storm_timer = paynt.utils.timer.Timer()
        storm_timer.start()
        result = belmc.check(self.spec_formulas[0], self.paynt_export)   # calls Storm
        storm_timer.stop()
        logger.info("Storm POMDP analysis completed")

        value = result.upper_bound if self.quotient.specification.optimality.minimizing else result.lower_bound
        self.belief_controller_size = self.get_belief_controller_size(result, self.paynt_fsc_size)

        print(f'-----------Storm-----------')
        print(f'Value = {value} | Time elapsed = {round(storm_timer.read(),1)}s | FSC size = {self.belief_controller_size}', flush=True)
        if self.get_result is not None:
            # TODO not important for the paper but it would be nice to have correct FSC here as well
            if self.storm_options == "overapp":
                #print(".....")
                #print(result.upper_bound)
                #print(result.lower_bound)
                pass
            # else:
                # print(f'FSC (dot) = {result.induced_mc_from_scheduler.to_dot()}\n', flush=True)

        # print(f'\nFSC (dot) = {result.induced_mc_from_scheduler.to_dot()}\n', flush=True)

        self.store_storm_result(result)

    # setup interactive Storm belief model checker
    def interactive_storm_setup(self):
        global belmc    # needs to be global for threading to work correctly
        options = self.get_interactive_options()
        belmc = stormpy.pomdp.BeliefExplorationModelCheckerDouble(self.pomdp, options)

    # start interactive belief model checker, this function is called only once to start the storm thread. To resume Storm computation 'interactive_storm_resume' is used
    def interactive_storm_start(self, storm_timeout):
        self.storm_thread = Thread(target=self.interactive_run, args=(belmc,))
        control_thread = Thread(target=self.interactive_control, args=(belmc, True, storm_timeout,))

        logger.info("Interactive Storm started")
        control_thread.start()
        self.storm_thread.start()

        control_thread.join()

    # resume interactive belief model checker, should be called only after belief model checker was previously started
    def interactive_storm_resume(self, storm_timeout):
        control_thread = Thread(target=self.interactive_control, args=(belmc, False, storm_timeout,))

        if self.storm_terminated:
            logger.info("Storm already terminated")
            return
        
        logger.info("Interactive Storm resumed")
        control_thread.start()

        control_thread.join()

    # terminate interactive belief model checker
    def interactive_storm_terminate(self):
        belmc.terminate_unfolding()
        self.storm_thread.join()

    # this function represents the storm thread in SAYNT
    def interactive_run(self, belmc):
        logger.info("starting Storm POMDP analysis")
        result = belmc.check(self.spec_formulas[0], self.paynt_export)   # calls Storm

        # to get here Storm exploration has to end either by constructing finite belief MDP or by outside termination
        self.storm_terminated = True

        if result.induced_mc_from_scheduler is not None:
            value = result.upper_bound if self.quotient.specification.optimality.minimizing else result.lower_bound
            size = self.get_belief_controller_size(result, self.paynt_fsc_size)

            print(f'-----------Storm----------- \
              \nValue = {value} | Time elapsed = {round(self.saynt_timer.read(),1)}s | FSC size = {size}\n', flush=True)

            self.store_storm_result(result)
            self.parse_results(self.quotient)
            self.update_data()

        logger.info("Storm POMDP analysis completed")

    # ensures correct execution of one loop of Storm exploration
    def interactive_control(self, belmc, start, storm_timeout):
        if belmc.has_converged():
            logger.info("Storm already terminated.")
            return

        # Update cut-off FSC values provided by PAYNT
        if not start:
            logger.info("Updating FSC values in Storm")
            belmc.set_fsc_values(self.paynt_export)
            belmc.continue_unfolding()

        # wait for Storm to start exploring
        while not belmc.is_exploring():
            if belmc.has_converged():
                break
            sleep(1)

        sleep(storm_timeout)
        if self.storm_terminated:
            logger.info("Storm terminated")
            return
        logger.info("Pausing Storm")
        belmc.pause_unfolding()

        # wait for the result to be constructed from the explored belief MDP
        while not belmc.is_result_ready():
            sleep(1)

        result = belmc.get_interactive_result()

        value = result.upper_bound if self.quotient.specification.optimality.minimizing else result.lower_bound
        size = self.get_belief_controller_size(result, self.paynt_fsc_size)

        print(f'-----------Storm----------- \
              \nValue = {value} | Time elapsed = {round(self.saynt_timer.read(),1)}s | FSC size = {size}\n', flush=True)

        self.store_storm_result(result)
        self.parse_results(self.quotient)
        self.update_data()

    ########
    # Different options for Storm below (would be nice to make this more succint)

    def get_cutoff_options(self, belief_states=100000):
        options = stormpy.pomdp.BeliefExplorationModelCheckerOptionsDouble(False, True)
        options.use_state_elimination_cutoff = False
        options.size_threshold_init = belief_states
        options.use_clipping = False
        return options

    def get_overapp_options(self, belief_states=20000000):
        options = stormpy.pomdp.BeliefExplorationModelCheckerOptionsDouble(True, False)
        options.use_state_elimination_cutoff = False
        options.size_threshold_init = belief_states
        options.use_clipping = False
        return options

    def get_refine_options(self, step_limit=0):
        options = stormpy.pomdp.BeliefExplorationModelCheckerOptionsDouble(False, True)
        options.use_state_elimination_cutoff = False
        options.size_threshold_init = 0
        options.size_threshold_factor = 2
        options.use_clipping = False
        options.gap_threshold_init = 0
        options.refine_precision = 0
        if step_limit > 0:
            options.refine_step_limit = step_limit
        options.refine = True
        return options

    def get_clip2_options(self):
        options = stormpy.pomdp.BeliefExplorationModelCheckerOptionsDouble(False, True)
        options.use_state_elimination_cutoff = False
        options.size_threshold_init = 0
        options.use_clipping = True
        options.clipping_grid_res = 2
        options.gap_threshold_init = 0
        return options

    def get_clip4_options(self):
        options = stormpy.pomdp.BeliefExplorationModelCheckerOptionsDouble(False, True)
        options.use_state_elimination_cutoff = False
        options.size_threshold_init = 0
        options.use_clipping = True
        options.clipping_grid_res = 4
        options.gap_threshold_init = 0
        return options

    def get_interactive_options(self):
        options = stormpy.pomdp.BeliefExplorationModelCheckerOptionsDouble(False, True)
        options.use_state_elimination_cutoff = False
        options.size_threshold_init = 0
        options.skip_heuristic_schedulers = False
        options.interactive_unfolding = True
        options.gap_threshold_init = 0
        options.refine = False
        options.cut_zero_gap = False
        if self.storm_options == "clip2":
            options.use_clipping = True
            options.clipping_grid_res = 2
        elif self.storm_options == "clip4":
            options.use_clipping = True
            options.clipping_grid_res = 4
        return options
    
    # End of options
    ########

    # computes over-approximation for the given POMDP
    # this can be used to compute bounds for POMDP abstraction
    # TODO discuss the best options for this use case
    @staticmethod
    def storm_pomdp_analysis(model, formulas):
        options = stormpy.pomdp.BeliefExplorationModelCheckerOptionsDouble(True, False)
        options.use_state_elimination_cutoff = False
        options.size_threshold_init = 1000000
        options.use_clipping = False
        options.exploration_time_limit = 60
        belmc = stormpy.pomdp.BeliefExplorationModelCheckerDouble(model, options)

        result = belmc.check(formulas[0], [])   # calls Storm

        return result
    
    # parse the current Storm and PAYNT results if they are available
    def parse_results(self, quotient):
        if self.latest_storm_result is not None:
            self.parse_storm_result(quotient)
        else:
            self.result_dict = {}
            self.result_dict_no_cutoffs = {}

        if self.latest_paynt_result is not None:
            self.parse_paynt_result(quotient)
        else:
            self.result_dict_paynt = {}
            
    # parse Storm results into a dictionary
    def parse_storm_result(self, quotient):
        # to make the code cleaner
        get_choice_label = self.latest_storm_result.induced_mc_from_scheduler.choice_labeling.get_labels_of_choice

        cutoff_epxloration = list(range(len(self.latest_storm_result.cutoff_schedulers)))
        finite_mem = False

        result = {x:[] for x in range(quotient.observations)}
        result_no_cutoffs = {x:[] for x in range(quotient.observations)}
        
        for state in self.latest_storm_result.induced_mc_from_scheduler.states:
            # TODO what if there were no labels in the model?
            if get_choice_label(state.id) == set():
                continue

            # parse non cut-off states
            if 'cutoff' not in state.labels and 'clipping' not in state.labels:
                for label in state.labels:
                    observation = None
                    if '[' in label:
                        # observation based on prism observables
                        observation = self.quotient.observation_labels.index(label)
                    elif 'obs_' in label:
                        # explicit observation index
                        _,observation = label.split('_')
                    if observation is not None:
                        observation = int(observation)
                        choice_label = list(get_choice_label(state.id))[0]
                        for index,action_label in enumerate(quotient.action_labels_at_observation[observation]):
                            if choice_label == action_label:
                                if index not in result[observation]:
                                    result[observation].append(index)
                                if index not in result_no_cutoffs[observation]:
                                    result_no_cutoffs[observation].append(index)
                                break
                        

            # parse cut-off states
            else:
                for label in state.labels:
                    if 'finite_mem' in label and not finite_mem:
                        finite_mem = True
                        self.parse_paynt_result(self.quotient)
                        for obs,actions in self.result_dict_paynt.items():
                            for action in actions:
                                if action not in result_no_cutoffs[obs]:
                                    result_no_cutoffs[obs].append(action)
                                if action not in result[obs]:
                                    result[obs].append(action)
                else:
                    if len(cutoff_epxloration) == 0:
                        continue
                    
                    # obtain what cut-off scheduler was used
                    if 'sched_' in list(get_choice_label(state.id))[0]:
                        _, scheduler_index = list(get_choice_label(state.id))[0].split('_')
    
                        if int(scheduler_index) not in cutoff_epxloration:
                            continue
    
                        scheduler = self.latest_storm_result.cutoff_schedulers[int(scheduler_index)]
    
                        for state in range(quotient.pomdp.nr_states):
    
                            choice_string = str(scheduler.get_choice(state).get_choice())
                            actions = self.parse_choice_string(choice_string)
    
                            observation = quotient.pomdp.get_observation(state)
    
                            for action in actions:
                                if action not in result[observation]:
                                    result[observation].append(action)
                        cutoff_epxloration.remove(int(scheduler_index))

        # removing unrestricted observations
        observations = list(result.keys())
        for obs in observations:
            if len(result[obs]) == 0:
                del result[obs]

            if len(result_no_cutoffs[obs]) == 0:
                del result_no_cutoffs[obs]

        self.result_dict = result    
        self.result_dict_no_cutoffs = result_no_cutoffs       
            

    # help function for cut-off parsing, returns list of actions for given choice_string
    # TODO bound to restrict some action if needed
    def parse_choice_string(self, choice_string, probability_bound=0):
        chars = '}{]['
        for c in chars:
            choice_string = choice_string.replace(c, '')
        choice_string = choice_string.strip(', ')
        choices = choice_string.split(',')

        result = []
        for choice in choices:
            probability, action = choice.split(':')
            # probability bound
            action = int(action.strip())
            result.append(action)

        return result

    # parse PAYNT result to a dictionart
    def parse_paynt_result(self, quotient):

        result = {x:[] for x in range(quotient.observations)}
        
        for hole in range(self.latest_paynt_result.num_holes):
            name = self.latest_paynt_result.hole_name(hole)
            if name.startswith('M'):
                continue
            name = name.strip('A()')
            obs = name.split(',')[0]
            observation = self.quotient.observation_labels.index(obs)

            option = self.latest_paynt_result.hole_options(hole)[0]
            if option not in result[observation]:
                result[observation].append(option)

        observations = list(result.keys())
        for obs in observations:
            if len(result[obs]) == 0:
                del result[obs]

        #logger.info("Result dictionary is based on result from PAYNT")
        self.result_dict_paynt = result

    # returns the main family that will be explored first
    # main family contains only the actions considered by respective FSC (most usually Storm result)
    def get_main_restricted_family(self, family, result_dict):

        if result_dict == {}:
            return family

        restricted_family = family.copy()
        # go through each observation of interest
        for obs in range(self.quotient.observations):
            for hole in self.quotient.observation_action_holes[obs]:

                if obs in result_dict.keys():
                    selected_actions = [action for action in family.hole_options(hole) if action in result_dict[obs]]
                else:
                    selected_actions = [family.hole_options(hole)[0]]

                if len(selected_actions) == 0:
                    return None

                restricted_family.hole_set_options(hole,selected_actions)

        logger.info("Main family based on data from Storm: reduced design space from {} to {}".format(family.size_or_order, restricted_family.size_or_order))

        return restricted_family


    # returns dictionary containing restrictions for easy creation of subfamilies
    # creating this restrictions list saves some memory compared to constructing all of the families
    # corresponding families are then created only when needed
    def get_subfamilies_restrictions(self, family, result_dict):

        if result_dict == {}:
            return {}

        subfamilies_restriction = []

        restricted_holes_list = []

        for observ in result_dict.keys():

            act_obs_holes = self.quotient.observation_action_holes[observ]
            restricted_holes_list.extend(act_obs_holes)

        for hole in restricted_holes_list:

            for obs_holes, index in zip(self.quotient.observation_action_holes, range(len(self.quotient.observation_action_holes))):
                if hole in obs_holes:
                    obs = index

            if len(result_dict[obs]) == self.quotient.actions_at_observation[obs]:
                continue

            restriction = [action for action in family.hole_options(hole) if action in result_dict[obs]]

            if len(restriction) == family.hole_num_options(hole):
                continue

            subfamilies_restriction.append({"hole": hole, "restriction": restriction})

        return subfamilies_restriction

    # constructs the families given by the restrictions list
    def get_subfamilies(self, restrictions, family):

        subfamilies = []

        for i,restriction in enumerate(restrictions):
            restricted_family = family.copy()

            actions = [action for action in family.hole_options(restriction["hole"]) if action not in restriction["restriction"]]
            if len(actions) == 0:
                actions = [family.hole_options(restriction["hole"])[0]]

            restricted_family.hole_set_options(restriction['hole'],actions)

            for j in range(i):
                restricted_family.hole_set_options(restrictions[j]['hole'],restrictions[j]["restriction"])

            subfamilies.append(restricted_family)

        return subfamilies

    # returns True if the current best FSC from Storm requires more memory
    # returns False otherwise
    def is_memory_needed(self):
        if len(self.memory_vector) == 0:
            return False

        memory_needed = False
        for obs in range(self.quotient.observations):
            if self.quotient.observation_memory_size[obs] < self.memory_vector[obs]:
                memory_needed = True
                break
        return memory_needed

    # update all of the important data structures according to the current results
    def update_data(self):

        if self.paynt_bounds is None and self.storm_bounds is None:
            return

        if self.paynt_bounds is None:
            self.is_storm_better = True
        elif self.storm_bounds is None:
            self.is_storm_better = False
        else:
            if self.quotient.specification.optimality.minimizing:
                if self.paynt_bounds <= self.storm_bounds:
                    self.is_storm_better = False
                else:
                    self.is_storm_better = True
            else:
                if self.paynt_bounds >= self.storm_bounds:
                    self.is_storm_better = False
                else:
                    self.is_storm_better = True


        if self.unfold_strategy_storm in ["storm", "paynt"]:
            for obs in range(self.quotient.observations):
                if obs in self.result_dict_no_cutoffs.keys():
                    self.memory_vector[obs] = len(self.result_dict_no_cutoffs[obs])
                else:
                    self.memory_vector[obs] = 1
        else:
            for obs in range(self.quotient.observations):
                if obs in self.result_dict.keys():
                    self.memory_vector[obs] = len(self.result_dict[obs])
                else:
                    self.memory_vector[obs] = 1

    def belief_controller_to_fsc(self, storm_result, paynt_fsc=None):

        belief_mc = storm_result.induced_mc_from_scheduler

        uses_fsc = False
        used_randomized_schedulers = {}
        paynt_cutoff_states = 0

        for state in belief_mc.states:
            for label in state.labels:
                if 'finite_mem' in label:
                    uses_fsc = True
                    paynt_cutoff_states += 1
                    continue
                elif 'sched_' in label:
                    _, scheduler_index = label.split('_')
                    if int(scheduler_index) in used_randomized_schedulers.keys():
                        continue
                    used_randomized_schedulers[int(scheduler_index)] = state.id
                    continue

        fsc_nodes = belief_mc.nr_states - paynt_cutoff_states + 1 # +1 for new initial node

        fsc_node = 1
        belief_mc_nodes_map = []
        for state in belief_mc.states:
            for label in state.labels:
                if 'finite_mem' in label:
                    belief_mc_nodes_map.append(None)
                    break
            else:
                belief_mc_nodes_map.append(fsc_node)
                fsc_node += 1

        assert fsc_nodes-1 == len([x for x in belief_mc_nodes_map if x is not None]), f"{fsc_nodes-1} != {len([x for x in belief_mc_nodes_map if x is not None])}"

        if uses_fsc:
            fsc_nodes += paynt_fsc.num_nodes
            first_fsc_node = belief_mc.nr_states - paynt_cutoff_states + 1

        result_fsc = paynt.quotient.fsc.FSC(fsc_nodes, self.quotient.observations, is_deterministic=False)

        action_labels = set()
        for labels in self.quotient.action_labels_at_observation:
            action_labels.update(labels)
        action_labels = list(action_labels)
        result_fsc.action_labels = action_labels

        result_fsc.observation_labels = self.quotient.observation_labels

        if paynt_fsc is not None and uses_fsc:
            paynt_fsc.make_stochastic()
            paynt_fsc_num_nodes = paynt_fsc.num_nodes
            paynt_fsc_action_function = paynt_fsc.action_function
            paynt_fsc_update_function = paynt_fsc.update_function
            new_fsc_update_function = []
            
            for node in range(paynt_fsc_num_nodes):
                new_fsc_update_function.append([])
                for obs in range(self.quotient.observations):
                    new_fsc_update_function[node].append({list(paynt_fsc_update_function[node][obs].keys())[0]+first_fsc_node:1.0})

            for node in range(paynt_fsc_num_nodes):
                result_fsc.action_function[node+first_fsc_node] = paynt_fsc_action_function[node]
                result_fsc.update_function[node+first_fsc_node] = new_fsc_update_function[node]
        
        # create the initial node
        init_belief_state = belief_mc.initial_states[0]
        actions = list(belief_mc.choice_labeling.get_labels_of_choice(init_belief_state))
        assert len(actions) == 1, "Belief MC has multiple labels for one action"
        action = actions[0]
        for label in belief_mc.labeling.get_labels_of_state(init_belief_state):
            succ_observation = None
            fsc_switch = None
            cutoff_switch = None
            for label in belief_mc.labeling.get_labels_of_state(init_belief_state):
                if '[' in label:
                    # observation based on prism observables
                    succ_observation = self.quotient.observation_labels.index(label)
                elif 'obs_' in label:
                    # explicit observation index
                    _,succ_observation = label.split('_')
                elif 'finite_mem' in label:
                    fsc_switch = int(label.split('_')[-1])
                if 'sched_' in label:
                    cutoff_switch = int(label.split('_')[1])
                if succ_observation is not None:
                    succ_observation = int(succ_observation)
            assert not((fsc_switch is not None) and (cutoff_switch is not None)), "Belief MC state has both FSC and Storm cutoff scheduler"
            assert succ_observation is not None, "Belief MC state has no observation"
            if fsc_switch is not None:
                result_fsc.action_function[0][succ_observation] = result_fsc.action_function[first_fsc_node+fsc_switch][succ_observation]
                result_fsc.update_function[0][succ_observation] = result_fsc.update_function[first_fsc_node+fsc_switch][succ_observation]
            elif cutoff_switch is not None:
                scheduler = storm_result.cutoff_schedulers[cutoff_switch]
                cutoff_node_id = belief_mc_nodes_map[used_randomized_schedulers[cutoff_switch]]
                for pomdp_state in range(self.quotient.pomdp.nr_states):
                    obs_index = self.quotient.pomdp.get_observation(pomdp_state)
                    if obs_index != succ_observation:
                        continue
                    choice = scheduler.get_choice(pomdp_state).get_choice().__str__()
                    choice = choice.replace('{','').replace('}','').replace('[','').replace(']','').replace(' ','').split(',')
                    for c in choice[:-1]:
                        prob, cutoff_action = c.split(':')
                        action_label = self.quotient.action_labels_at_observation[succ_observation][int(cutoff_action)]
                        action_index = result_fsc.action_labels.index(action_label)
                        if result_fsc.action_function[node_id][succ_observation] is None:
                            result_fsc.action_function[node_id][succ_observation] = {action_index:float(prob)}
                        else:
                            result_fsc.action_function[node_id][succ_observation][action_index] = float(prob)
                    break
                result_fsc.update_function[0][succ_observation] = {cutoff_node_id: 1.0}
            else:
                if action == "loop":
                    result_fsc.action_function[0][succ_observation] = {0: 1.0}
                else:
                    result_fsc.action_function[0][succ_observation] = {result_fsc.action_labels.index(action): 1.0}
                result_fsc.update_function[0][succ_observation] = {belief_mc_nodes_map[init_belief_state]: 1.0}

        for state in belief_mc.states:
            node_id = belief_mc_nodes_map[state.id]
            if node_id is None:
                continue
            elif 'cutoff' in state.labels: # Storm cutoff schedulers
                for label in state.labels:
                    if 'sched_' in label:
                        _, scheduler_index = label.split('_')
                        break
                else:
                    continue
                scheduler = storm_result.cutoff_schedulers[int(scheduler_index)]
                for obs in range(self.quotient.observations):
                    result_fsc.update_function[node_id][obs] = {node_id: 1.0}
                processed_obs = []
                for pomdp_state in range(self.quotient.pomdp.nr_states):
                    obs_index = self.quotient.pomdp.get_observation(pomdp_state)
                    if obs_index in processed_obs:
                        continue
                    processed_obs.append(obs_index)
                    choice = scheduler.get_choice(pomdp_state).get_choice().__str__()
                    choice = choice.replace('{','').replace('}','').replace('[','').replace(']','').replace(' ','').split(',')
                    for c in choice[:-1]:
                        prob, action = c.split(':')
                        action = int(action)
                        action_label = self.quotient.action_labels_at_observation[obs_index][action]
                        action_index = result_fsc.action_labels.index(action_label)
                        if result_fsc.action_function[node_id][obs_index] is None:
                            result_fsc.action_function[node_id][obs_index] = {action_index:float(prob)}
                        else:
                            result_fsc.action_function[node_id][obs_index][action_index] = float(prob)
            elif '__extra' in state.labels or 'target' in state.labels: # basically target states so just loop with everything
                for obs in range(self.quotient.observations):
                    first_action_in_obs = self.quotient.action_labels_at_observation[obs][0] # this ensures the looping action is available
                    result_fsc.action_function[node_id][obs] = {action_labels.index(first_action_in_obs):1.0}
                    result_fsc.update_function[node_id][obs] = {node_id: 1.0}
            else: # normal belief mc states
                successors = []
                for transition in belief_mc.transition_matrix.row_iter(state.id, state.id):
                    successors.append(transition.column)
                for succ in successors:
                    actions = list(belief_mc.choice_labeling.get_labels_of_choice(succ))
                    assert len(actions) == 1, "Belief MC has multiple labels for one action"
                    action = actions[0]
                    succ_observation = None
                    fsc_switch = None
                    cutoff_switch = None
                    for label in belief_mc.labeling.get_labels_of_state(succ):
                        if '[' in label:
                            # observation based on prism observables
                            succ_observation = self.quotient.observation_labels.index(label)
                        elif 'obs_' in label:
                            # explicit observation index
                            _,succ_observation = label.split('_')
                        elif 'finite_mem' in label:
                            fsc_switch = int(label.split('_')[-1])
                        if 'sched_' in label:
                            cutoff_switch = int(label.split('_')[1])
                        if succ_observation is not None:
                            succ_observation = int(succ_observation)
                    assert not((fsc_switch is not None) and (cutoff_switch is not None)), "Belief MC state has both FSC and Storm cutoff scheduler"
                    assert succ_observation is not None, "Belief MC state has no observation"
                    if fsc_switch is not None:
                        result_fsc.action_function[node_id][succ_observation] = result_fsc.action_function[first_fsc_node+fsc_switch][succ_observation]
                        result_fsc.update_function[node_id][succ_observation] = result_fsc.update_function[first_fsc_node+fsc_switch][succ_observation]
                    elif cutoff_switch is not None:
                        scheduler = storm_result.cutoff_schedulers[cutoff_switch]
                        cutoff_node_id = belief_mc_nodes_map[used_randomized_schedulers[cutoff_switch]]
                        for pomdp_state in range(self.quotient.pomdp.nr_states):
                            obs_index = self.quotient.pomdp.get_observation(pomdp_state)
                            if obs_index != succ_observation:
                                continue
                            choice = scheduler.get_choice(pomdp_state).get_choice().__str__()
                            choice = choice.replace('{','').replace('}','').replace('[','').replace(']','').replace(' ','').split(',')
                            for c in choice[:-1]:
                                prob, cutoff_action = c.split(':')
                                action_label = self.quotient.action_labels_at_observation[succ_observation][int(cutoff_action)]
                                action_index = result_fsc.action_labels.index(action_label)
                                if result_fsc.action_function[node_id][succ_observation] is None:
                                    result_fsc.action_function[node_id][succ_observation] = {action_index:float(prob)}
                                else:
                                    result_fsc.action_function[node_id][succ_observation][action_index] = float(prob)
                            break
                        result_fsc.update_function[node_id][succ_observation] = {cutoff_node_id: 1.0}
                    else:
                        if action == "loop":
                            first_action_in_obs = self.quotient.action_labels_at_observation[succ_observation][0] # this ensures the looping action is available
                            result_fsc.action_function[node_id][succ_observation] = {action_labels.index(first_action_in_obs): 1.0}
                        else:
                            result_fsc.action_function[node_id][succ_observation] = {result_fsc.action_labels.index(action): 1.0}
                        result_fsc.update_function[node_id][succ_observation] = {belief_mc_nodes_map[succ]: 1.0}

        logger.info(f"constructed FSC with {result_fsc.num_nodes} nodes")

        return result_fsc


    
    # Computes the size of the controller for belief MC
    # if it uses FSC cutoffs assignment should be provided
    # FORMULA: E + 2*T + size(Fc)
    # E - number of non-frontier states (non cutoff states)
    # T - number of transitions
    # Fc - used cut-off schedulers
    def get_belief_controller_size(self, storm_result, paynt_fsc_size=None):

        belief_mc = storm_result.induced_mc_from_scheduler

        non_frontier_states = 0
        uses_fsc = False
        used_randomized_schedulers = []

        fsc_size = 0
        randomized_schedulers_size = 0

        for state in belief_mc.states:
            if 'cutoff' not in state.labels:
                non_frontier_states += 1
            elif 'finite_mem' in state.labels and not uses_fsc:
                uses_fsc = True
            else:
                for label in state.labels:
                    if 'sched_' in label:
                        _, scheduler_index = label.split('_')
                        if int(scheduler_index) in used_randomized_schedulers:
                            continue
                        used_randomized_schedulers.append(int(scheduler_index))

        if uses_fsc:
            # Compute the size of FSC
            if paynt_fsc_size:
                fsc_size = paynt_fsc_size

        for index in used_randomized_schedulers:
            observation_actions = {x:[] for x in range(self.quotient.observations)}
            rand_scheduler = storm_result.cutoff_schedulers[index]
            for state in range(self.quotient.pomdp.nr_states):
                choice_string = str(rand_scheduler.get_choice(state).get_choice())
                actions = self.parse_choice_string(choice_string)
                observation = self.quotient.pomdp.get_observation(state)
                for action in actions:
                    if action not in observation_actions[observation]:
                        observation_actions[observation].append(action)
            randomized_schedulers_size += sum(list([len(support) for support in observation_actions.values()])) * 3

        result_size = non_frontier_states + belief_mc.nr_transitions + fsc_size + randomized_schedulers_size

        return result_size



import stormpy

from .statistic import Statistic
import paynt.synthesizer.synthesizer_ar
import paynt.synthesizer.synthesizer_hybrid
import paynt.synthesizer.synthesizer_ar_storm
import paynt.synthesizer.synthesizer

import paynt.quotient.quotient
import paynt.quotient.pomdp
import paynt.utils.timer

import paynt.verification.property

from threading import Thread
from queue import Queue
import time

import logging
logger = logging.getLogger(__name__)


class SynthesizerPomdp:

    # If true explore only the main family
    incomplete_exploration = False

    def __init__(self, quotient, method, storm_control):
        self.quotient = quotient
        self.synthesizer = None
        if method == "ar":
            self.synthesizer = paynt.synthesizer.synthesizer_ar.SynthesizerAR
        elif method == "hybrid":
            self.synthesizer = paynt.synthesizer.synthesizer_hybrid.SynthesizerHybrid
        self.total_iters = 0

        self.storm_control = storm_control
        if storm_control is not None:
            self.storm_control.quotient = self.quotient
            self.storm_control.pomdp = self.quotient.pomdp
            self.storm_control.spec_formulas = self.quotient.specification.stormpy_formulae()
            self.synthesis_terminate = False
            self.synthesizer = paynt.synthesizer.synthesizer_ar_storm.SynthesizerARStorm # SAYNT only works with abstraction refinement
            if self.storm_control.iteration_timeout is not None:
                self.saynt_timer = paynt.utils.timer.Timer()
                self.synthesizer.saynt_timer = self.saynt_timer
                self.storm_control.saynt_timer = self.saynt_timer

    def synthesize(self, family=None, print_stats=True):
        if family is None:
            family = self.quotient.family
        synthesizer = self.synthesizer(self.quotient)
        family.constraint_indices = self.quotient.family.constraint_indices
        assignment = synthesizer.synthesize(family, keep_optimum=True, print_stats=print_stats)
        iters_mdp = synthesizer.stat.iterations_mdp if synthesizer.stat.iterations_mdp is not None else 0
        self.total_iters += iters_mdp
        return assignment

    def unfold_and_synthesize(self, mem_size, unfold_storm, unfold_imperfect_only=True):
        paynt.quotient.pomdp.PomdpQuotient.current_family_index = mem_size

        # unfold memory according to the best result
        if not unfold_storm:
            logger.info("Synthesizing optimal k={} controller ...".format(mem_size) )
            if unfold_imperfect_only:
                self.quotient.set_imperfect_memory_size(mem_size)
            else:
                self.quotient.set_global_memory_size(mem_size)
        else:
            if mem_size > 1:
                obs_memory_dict = {}
                if self.storm_control.is_storm_better:
                    # Storm's result is better and it needs memory
                    if self.storm_control.is_memory_needed():
                        obs_memory_dict = self.storm_control.memory_vector
                        logger.info(f'Added memory nodes to match Storm data')
                    else:
                        if self.storm_control.unfold_cutoff:
                            # consider the cut-off schedulers actions when updating memory
                            result_dict = self.storm_control.result_dict
                        else:
                            # only consider the induced DTMC without cut-off states
                            result_dict = self.storm_control.result_dict_no_cutoffs
                        for obs in range(self.quotient.observations):
                            if obs in result_dict:
                                obs_memory_dict[obs] = self.quotient.observation_memory_size[obs] + 1
                            else:
                                obs_memory_dict[obs] = self.quotient.observation_memory_size[obs]
                        logger.info(f'Added memory nodes for observations based on Storm data')
                else:
                    for obs in range(self.quotient.observations):
                        if self.quotient.observation_states[obs]>1:
                            obs_memory_dict[obs] = self.quotient.observation_memory_size[obs] + 1
                        else:
                            obs_memory_dict[obs] = 1
                    logger.info(f'Increased memory in all imperfect observation')
                self.quotient.set_memory_from_dict(obs_memory_dict)

        family = self.quotient.family

        # if Storm's result is better, use it to obtain main family that considers only the important actions
        if self.storm_control.is_storm_better:
            if self.storm_control.use_cutoffs:
                # consider the cut-off schedulers actions
                result_dict = self.storm_control.result_dict
            else:
                # only consider the induced DTMC actions without cut-off states
                result_dict =self.storm_control.result_dict_no_cutoffs
            main_family = self.storm_control.get_main_restricted_family(family,result_dict)
            subfamily_restrictions = []
            if not self.storm_control.incomplete_exploration:
                subfamily_restrictions = self.storm_control.get_subfamilies_restrictions(family, result_dict)
            subfamilies = self.storm_control.get_subfamilies(subfamily_restrictions, family)
        # if PAYNT is better continue normally
        else:
            main_family = family
            subfamilies = []

        self.synthesizer.subfamilies_buffer = subfamilies
        self.synthesizer.main_family = main_family

        assignment = self.synthesize(family)
        return assignment

    # iterative strategy using Storm analysis to enhance the synthesis
    def strategy_iterative_storm(self, unfold_imperfect_only, unfold_storm=True):
        '''
        @param unfold_imperfect_only if True, only imperfect observations will be unfolded
        '''
        mem_size = paynt.quotient.pomdp.PomdpQuotient.initial_memory_size
        self.synthesizer.storm_control = self.storm_control

        while True:
            assignment = self.unfold_and_synthesize(mem_size,unfold_storm)
            if assignment is not None:
                self.storm_control.latest_paynt_result = assignment
                self.storm_control.paynt_export = self.quotient.extract_policy(assignment)
                self.storm_control.paynt_bounds = self.quotient.specification.optimality.optimum
                self.storm_control.paynt_fsc_size = self.quotient.policy_size(self.storm_control.latest_paynt_result)
                self.storm_control.latest_paynt_result_fsc = self.quotient.assignment_to_fsc(self.storm_control.latest_paynt_result)
            self.storm_control.update_data()

            if self.synthesis_terminate:
                break

            mem_size += 1


    def print_synthesized_controllers(self):
        hline = "\n------------------------------------\n"
        print(hline)
        print("PAYNT results: ")
        print(self.storm_control.paynt_bounds)
        print("controller size: {}".format(self.storm_control.paynt_fsc_size))
        print()
        print("Storm results: ")
        print(self.storm_control.storm_bounds)
        print("controller size: {}".format(self.storm_control.belief_controller_size))
        print(hline)


    def iterative_storm_loop(self, timeout, paynt_timeout, storm_timeout, iteration_limit=0):
        ''' Main SAYNT loop. '''
        self.interactive_queue = Queue()
        self.synthesizer.s_queue = self.interactive_queue
        self.storm_control.interactive_storm_setup()
        iteration = 1
        paynt_thread = Thread(target=self.strategy_iterative_storm, args=(True, self.storm_control.unfold_storm))

        iteration_timeout = time.time() + timeout

        self.saynt_timer.start()
        while True:
            if iteration == 1:
                paynt_thread.start()
            else:
                self.interactive_queue.put("resume")

            logger.info("Timeout for PAYNT started")

            time.sleep(paynt_timeout)
            self.interactive_queue.put("timeout")

            while not self.interactive_queue.empty():
                time.sleep(0.1)

            if iteration == 1:
                self.storm_control.interactive_storm_start(storm_timeout)
            else:
                self.storm_control.interactive_storm_resume(storm_timeout)

            # compute sizes of controllers
            assert self.storm_control.latest_storm_result is not None
            self.storm_control.belief_controller_size = self.storm_control.get_belief_controller_size(self.storm_control.latest_storm_result, self.storm_control.paynt_fsc_size)

            self.print_synthesized_controllers()

            if time.time() > iteration_timeout or iteration == iteration_limit:
                break

            iteration += 1

        self.interactive_queue.put("terminate")
        self.synthesis_terminate = True
        paynt_thread.join()

        self.storm_control.interactive_storm_terminate()

        self.saynt_timer.stop()

    # run PAYNT POMDP synthesis with a given timeout
    def run_synthesis_timeout(self, timeout):
        self.interactive_queue = Queue()
        self.synthesizer.s_queue = self.interactive_queue
        paynt_thread = Thread(target=self.strategy_iterative_storm, args=(True, False))
        iteration_timeout = time.time() + timeout
        paynt_thread.start()

        while True:
            if time.time() > iteration_timeout:
                break

            time.sleep(1)

        self.interactive_queue.put("pause")
        self.interactive_queue.put("terminate")
        self.synthesis_terminate = True
        paynt_thread.join()


    # PAYNT POMDP synthesis that uses pre-computed results from Storm as guide
    def strategy_storm(self, unfold_imperfect_only, unfold_storm=True):
        '''
        @param unfold_imperfect_only if True, only imperfect observations will be unfolded
        '''
        mem_size = paynt.quotient.pomdp.PomdpQuotient.initial_memory_size
        self.synthesizer.storm_control = self.storm_control

        while True:
            if self.storm_control.is_storm_better == False:
                self.storm_control.parse_results(self.quotient)
            assignment = self.unfold_and_synthesize(mem_size,unfold_storm)
            if assignment is not None:
                self.storm_control.latest_paynt_result = assignment
                self.storm_control.paynt_export = self.quotient.extract_policy(assignment)
                self.storm_control.paynt_bounds = self.quotient.specification.optimality.optimum

            self.storm_control.update_data()
            mem_size += 1


    def strategy_iterative(self, unfold_imperfect_only):
        '''
        @param unfold_imperfect_only if True, only imperfect observations will be unfolded
        '''
        mem_size = paynt.quotient.pomdp.PomdpQuotient.initial_memory_size
        opt = self.quotient.specification.optimality.optimum
        while True:

            logger.info("Synthesizing optimal k={} controller ...".format(mem_size) )
            if unfold_imperfect_only:
                self.quotient.set_imperfect_memory_size(mem_size)
            else:
                self.quotient.set_global_memory_size(mem_size)

            self.synthesize(self.quotient.family)

            opt_old = opt
            opt = self.quotient.specification.optimality.optimum

            # finish if optimum has not been improved
            # if opt_old == opt and opt is not None:
            #     break
            mem_size += 1

            #break

    def export_fsc(self, export_filename_base):

        fsc_json = None
        if self.storm_control.saynt_fsc is not None:
            fsc_json = self.storm_control.saynt_fsc.__str__()
        elif self.storm_control.latest_paynt_result_fsc is not None:
            fsc_json = self.storm_control.latest_paynt_result_fsc.__str__()
        else:
            # TODO add export option for pure PAYNT synthesis
            pass
        
        assert fsc_json is not None, "No FSC to export"

        with open(export_filename_base + ".fsc.json", "w") as f:
            f.write(fsc_json)

        logger.info(f"Exported FSC to {export_filename_base}.fsc.json")

    def run(self, optimum_threshold=None):
        if self.storm_control is None:
            # Pure PAYNT POMDP synthesis
            self.strategy_iterative(unfold_imperfect_only=True)
        else:
            # SAYNT
            logger.info("Storm POMDP option enabled")
            logger.info("Storm settings: iterative - {}, get_storm_result - {}, storm_options - {}, prune_storm - {}, unfold_strategy - {}, use_storm_cutoffs - {}".format(
                        (self.storm_control.iteration_timeout, self.storm_control.paynt_timeout, self.storm_control.storm_timeout), self.storm_control.get_result,
                        self.storm_control.storm_options, self.storm_control.incomplete_exploration, (self.storm_control.unfold_storm, self.storm_control.unfold_cutoff), self.storm_control.use_cutoffs
            ))
            # start SAYNT
            if self.storm_control.iteration_timeout is not None:
                self.iterative_storm_loop(timeout=self.storm_control.iteration_timeout,
                                        paynt_timeout=self.storm_control.paynt_timeout,
                                        storm_timeout=self.storm_control.storm_timeout,
                                        iteration_limit=0)
            # run PAYNT for a time given by 'self.storm_control.get_result' and then run Storm using the best computed FSC at cut-offs
            elif self.storm_control.get_result is not None:
                if self.storm_control.get_result:
                    self.run_synthesis_timeout(self.storm_control.get_result)
                self.storm_control.run_storm_analysis()
            # run Storm and then use the obtained result to enhance PAYNT synthesis
            else:
                self.storm_control.get_storm_result()
                self.strategy_storm(unfold_imperfect_only=True, unfold_storm=self.storm_control.unfold_storm)

            self.print_synthesized_controllers()

        if paynt.synthesizer.synthesizer.Synthesizer.export_synthesis_filename_base is not None:
            self.export_fsc(paynt.synthesizer.synthesizer.Synthesizer.export_synthesis_filename_base)



import stormpy
import payntbind

import paynt.family.family
import paynt.synthesizer.synthesizer

import paynt.quotient.quotient
import paynt.verification.property_result
from paynt.verification.property import Property
import paynt.utils.timer

import paynt.family.smt
import paynt.synthesizer.conflict_generator.dtmc
import paynt.synthesizer.conflict_generator.mdp


import logging
logger = logging.getLogger(__name__)

# disable logging when importing graphviz to suppress warnings
logging.disable(logging.CRITICAL)
import graphviz
logging.disable(logging.NOTSET)


def policies_are_compatible(policy1, policy2):
    policy1,policy1_mask = policy1
    policy2,_ = policy2
    for state in policy1_mask:
        a1 = policy1[state]
        a2 = policy2[state]
        if a2 is not None and a1 != a2:
            return False
    return True

def merge_policies(policy1, policy2):
    '''
    Attempt to merge multiple policies into one.
    :returns one policy or None if some policies were incompatible
    '''
    if not policies_are_compatible(policy1,policy2):
        return None
    policy1,_ = policy1
    policy2,_ = policy2
    policy = [a1 or policy2[state] for state,a1 in enumerate(policy1)]
    mask = [state for state,action in enumerate(policy) if action is not None]
    return (policy,mask)

def merge_policies_exclusively(policy1, policy2):
    policy1,_ = policy1
    policy2,_ = policy2
    policy12 = policy1.copy()
    policy21 = policy2.copy()
    for state,a1 in enumerate(policy1):
        a2 = policy2[state]
        if a1 is None:
            policy12[state] = a2
        if a2 is None:
            policy21[state] = a1
    return policy12,policy21


class PolicyTreeNode:

    def __init__(self, family):
        self.family = family
        
        self.splitter = None
        self.suboptions = []
        self.child_nodes = []

        self.sat = None
        self.policy_index = None

    @property
    def is_leaf(self):
        return self.sat is not None
    
    def num_nodes(self):
        num = 1
        for child in self.child_nodes:
            num += child.num_nodes()
        return num

    def num_leaves(self):
        num = 1 if self.is_leaf else 0
        for child in self.child_nodes:
            num += child.num_leaves()
        return num

    def split(self, splitter, suboptions, subfamilies):
        self.splitter = splitter
        self.suboptions = suboptions
        self.child_nodes = []
        for subfamily in subfamilies:
            child_node = PolicyTreeNode(subfamily)
            self.child_nodes.append(child_node)

    def double_check(self, quotient, prop, policies):
        assert self.sat is not None
        quotient.build(self.family)
        if self.sat is False:
            result = self.family.mdp.model_check_property(prop)
            assert not result.sat
        else:
            SynthesizerPolicyTree.double_check_policy(quotient, self.family, prop, policies[self.policy_index][0])


    def merge_children_indices(self, indices):
        if len(indices) <= 1:
            return
        target = indices[0]
        for j in reversed(indices[1:]):
            self.suboptions[target] += self.suboptions[j]
            self.child_nodes[target].family.hole_set_options(self.splitter,self.suboptions[target])
            self.suboptions.pop(j)
            self.child_nodes.pop(j)

        if len(self.child_nodes) > 1:
            return
        # a single child remains, can be merged into parent
        child_node = self.child_nodes[0]
        self.sat = child_node.sat
        self.policy_index = child_node.policy_index
        self.splitter = None
        self.suboptions = []
        self.child_nodes = []

    def merge_children_sat(self):
        indices = [i for i,child in enumerate(self.child_nodes) if child.sat is True]
        self.merge_children_indices(indices)

    def merge_children_having_same_solution(self):
        if self.is_leaf:
            return

        # merge UNSAT children
        indices = [i for i,child in enumerate(self.child_nodes) if child.sat is False]
        self.merge_children_indices(indices)

        # merge children having the same policy
        i = 0
        while i < len(self.child_nodes):
            child1 = self.child_nodes[i]
            if child1.sat is not True:
                i += 1
                continue

            join_to_i = [i]
            # collect other children to merge to i
            for j in range(i+1,len(self.child_nodes)):
                child2 = self.child_nodes[j]
                if child2.policy_index == child1.policy_index:
                    join_to_i.append(j)
            self.merge_children_indices(join_to_i)
            i += 1

    def make_policies_compatible(quotient, prop, node1, node2, policies):
        policy1 = policies[node1.policy_index]
        policy2 = policies[node2.policy_index]
        policy = merge_policies(policy1,policy2)
        if policy is not None:
            return policy
        
        policy12,policy21 = merge_policies_exclusively(policy1,policy2)

        # try policy1 for family2
        policy,mdp = quotient.fix_and_apply_policy_to_family(node2.family, policy12)
        policy_result = mdp.model_check_property(prop, alt=True)
        PolicyTreeNode.mdps_model_checked += 1
        if policy_result.sat:
            return policy

        # try policy2 for family1
        policy,mdp = quotient.fix_and_apply_policy_to_family(node1.family, policy21)
        policy_result = mdp.model_check_property(prop, alt=True)
        PolicyTreeNode.mdps_model_checked += 2
        if policy_result.sat:
            return policy

        # neither fits
        return None

    def merge_children_having_compatible_policies(self, quotient, prop, policies):
        if self.is_leaf:
            return
        i = 0
        while i < len(self.child_nodes):
            child1 = self.child_nodes[i]
            if child1.sat is not True:
                i += 1
                continue

            join_to_i = [i]
            # collect other children to merge to i
            for j in range(i+1,len(self.child_nodes)):
                child2 = self.child_nodes[j]
                if child2.sat is not True:
                    continue
                policy = PolicyTreeNode.make_policies_compatible(quotient,prop,child1,child2,policies)
                if policy is None:
                    continue
                # nodes can be merged
                policies[child1.policy_index] = policy
                policies[child2.policy_index] = None
                join_to_i.append(j)
            
            self.merge_children_indices(join_to_i)
            i += 1

    def skip_redundant_children(self):
        ''' Adopt grandchildren of each child that uses the same splitter as self. '''
        if self.splitter is None:
            return
        suboptions = []
        child_nodes = []
        for child_index,child in enumerate(self.child_nodes):
            if child.splitter != self.splitter:
                suboptions.append(self.suboptions[child_index])
                child_nodes.append(self.child_nodes[child_index])
            else:
                for grandchild_index,grandchild in enumerate(child.child_nodes):
                    suboptions.append(child.suboptions[grandchild_index])
                    child_nodes.append(grandchild)
        self.suboptions = suboptions
        self.child_nodes = child_nodes



    @property
    def node_id(self):
        return str(self.family).replace(' ','').replace(':','=')

    def add_nodes_to_graphviz_tree(self, graphviz_tree):
        node_label = ""
        if self.sat is False:
            node_label = "∅"
            # node_label = "X"
        elif self.sat is True:
            # node_label = "✓"
            node_label = f"p{self.policy_index}"
        graphviz_tree.node(self.node_id, label=node_label, shape="ellipse", width="0.15", height="0.15")
        # enumerating in reverse to print policies in ascending order, from left to right
        for child in reversed(self.child_nodes):
            child.add_nodes_to_graphviz_tree(graphviz_tree)

    def add_edges_to_graphviz_tree(self, graphviz_tree):
        if self.splitter is None:
            return
        splitter_name = self.family.hole_name(self.splitter)
        for index,child in enumerate(self.child_nodes):
            edge_label = self.family.hole_options_to_string(self.splitter,self.suboptions[index])
            graphviz_tree.edge(self.node_id,child.node_id,label=edge_label)
            child.add_edges_to_graphviz_tree(graphviz_tree)



class PolicyTree:

    def __init__(self, family):
        self.root = PolicyTreeNode(family)
        self.policies = []

    def new_policy(self, policy):
        policy_index = len(self.policies)
        mask = [state for state,action in enumerate(policy) if action is not None]
        self.policies.append( (policy,mask) )
        return policy_index

    def collect_all(self):
        node_queue = [self.root]
        all_nodes = []
        while node_queue:
            node = node_queue.pop(0)
            all_nodes.append(node)
            node_queue += node.child_nodes
        return all_nodes
    
    def collect_leaves(self):
        node_queue = [self.root]
        leaves = []
        while node_queue:
            node = node_queue.pop(0)
            if node.is_leaf:
                leaves.append(node)
            else:
                node_queue = node_queue + node.child_nodes
        return leaves

    def collect_nonleaves(self):
        node_queue = [self.root]
        nonleaves = []
        while node_queue:
            node = node_queue.pop(0)
            if not node.is_leaf:
                nonleaves.append(node)
                node_queue = node_queue + node.child_nodes
        return nonleaves

    def collect_sat(self):
        node_queue = [self.root]
        sat = []
        while node_queue:
            node = node_queue.pop(0)
            if node.sat:
                sat.append(node)
            else:
                node_queue += node.child_nodes
        return sat


    def double_check(self, quotient, prop):
        leaves = self.collect_leaves()
        logger.info("double-checking {} families...".format(len(leaves)))
        for leaf in leaves:
            leaf.double_check(quotient,prop,self.policies)
        logger.info("all solutions are OK")

    
    def count_diversity(self):
        from collections import defaultdict
        children_stats = defaultdict(int)
        for node in self.collect_all():
            if node.is_leaf:
                continue
            with_policy = len([node for node in node.child_nodes if node.sat])
            with_none = len([node for node in node.child_nodes if node.policy is None])
            with_false = len([node for node in node.child_nodes if node.policy==False])
            children_stats[(with_policy,with_none,with_false)] += 1
        return children_stats
    

    def print_stats(self):
        members_total = self.root.family.size
        num_policies = len(self.policies)
        
        members_satisfied = 0
        num_leaves_singleton = 0
        leaves = self.collect_leaves()
        for node in leaves:
            if node.family.size==1:
                num_leaves_singleton += 1
            if node.sat:
                members_satisfied += node.family.size
        satisfied_percentage = round(members_satisfied/members_total*100,0)
        members_unsatisfied = members_total-members_satisfied

        num_nodes = len(self.collect_all())
        num_leaves = len(leaves)
        num_leaves_solvable = len(self.collect_sat())
        num_leaves_unsolvable = num_leaves-num_leaves_solvable
        if num_leaves_solvable > 0:
            leaf_solvable_avg = round(members_satisfied / num_leaves_solvable,1)
        else:
            leaf_solvable_avg = "NA"

        if num_leaves_unsolvable > 0:
            leaf_unsolvable_avg = round(members_unsatisfied / num_leaves_unsolvable,1)
        else:
            leaf_unsolvable_avg = "NA"

        print("--------------------")
        print("Policy tree summary:")
        print("found {} satisfying {} for {}/{} family members ({}%)".format(
            num_policies, "policy" if num_policies==1 else "policies", members_satisfied,members_total,satisfied_percentage))
        print("policy tree has {} nodes, {} of them are leaves:".format(num_nodes, num_leaves))
        print("\t  solvable leaves: {} (avg.size: {})".format(num_leaves_solvable,leaf_solvable_avg))
        print("\tunsolvable leaves: {} (avg.size: {})".format(num_leaves_unsolvable,leaf_unsolvable_avg))
        print("\t singleton leaves: {}".format(num_leaves_singleton))

        print("--------------------")

    def discard_unused_policies(self):
        policy_old_to_new  = [None for _ in self.policies]
        num_policies = 0
        for policy_index,policy in enumerate(self.policies):
            if policy is not None:
                policy_old_to_new[policy_index] = num_policies
                num_policies += 1
        self.policies = [policy for policy in self.policies if policy is not None]
        assert num_policies == len(self.policies)
        for leaf in self.collect_sat():
            leaf.policy_index = policy_old_to_new[leaf.policy_index]
            assert leaf.policy_index is not None

    def merge_compatible_policies(self, policy_indices):
        policy_old_to_new_map = [policy_index for policy_index,_ in enumerate(self.policies)]

        for policy1_index_index,policy1_index in enumerate(policy_indices):
            policy1 = self.policies[policy1_index]
            if policy1 is None:
                continue
            for policy2_index in policy_indices[policy1_index_index+1:]:
                policy2 = self.policies[policy2_index]
                if policy2 is None:
                    continue
                policy = merge_policies(policy1,policy2)
                if policy is None:
                    continue
                # store updated policy
                self.policies[policy1_index] = policy
                policy1 = policy
                # discard irrelevant policy
                policy_old_to_new_map[policy2_index] = policy1_index
                self.policies[policy2_index] = None
        
        return policy_old_to_new_map
    
    def postprocess(self, quotient, prop):

        postprocessing_timer = paynt.utils.timer.Timer()
        postprocessing_timer.start()
        logger.info("post-processing the policy tree...")

        logger.info("merging SAT siblings solved by non-exclusively compatible policies...")
        PolicyTreeNode.mdps_model_checked = 0
        nodes_before = self.root.num_nodes()
        for node in reversed(self.collect_all()):
            node.merge_children_having_compatible_policies(quotient, prop, self.policies)
        self.discard_unused_policies()
        nodes_removed = nodes_before - self.root.num_nodes()
        logger.info("additional {} MDPs were model checked".format(PolicyTreeNode.mdps_model_checked))
        logger.info("removed {} nodes".format(nodes_removed))

        logger.info("merging all exclusively compatible policies...")
        policies_before = len(self.policies)
        policy_indices = [index for index,_ in enumerate(self.policies)]
        policy_old_to_new_map = self.merge_compatible_policies(policy_indices)
        for leaf in self.collect_sat():
            leaf.policy_index = policy_old_to_new_map[leaf.policy_index]
        self.discard_unused_policies()
        policies_removed = policies_before - len(self.policies)
        logger.info("removed {} policies".format(policies_removed))

        logger.info("reducing tree height...")
        nodes_before = self.root.num_nodes()
        for node in reversed(self.collect_nonleaves()):
            node.skip_redundant_children()
        nodes_removed = nodes_before - self.root.num_nodes()
        logger.info("removed {} nodes".format(nodes_removed))

        logger.info("merging siblings that have the same solution...")
        nodes_before = self.root.num_nodes()
        for node in reversed(self.collect_nonleaves()):
            node.merge_children_having_same_solution()
        nodes_removed = nodes_before - self.root.num_nodes()
        logger.info("removed {} nodes".format(nodes_removed))

        postprocessing_timer.stop()
        time = int(postprocessing_timer.read())
        logger.debug(f"postprocessing took {time} s")
        return time

    
    def extract_policies(self, quotient):
        return {
            f"p{policy_index}" : quotient.policy_to_state_valuation_actions(policy)
            for policy_index,policy in enumerate(self.policies)
        }

    def extract_policy_tree(self, quotient):
        logging.getLogger("graphviz").setLevel(logging.WARNING)
        logging.getLogger("graphviz.sources").setLevel(logging.ERROR)
        graphviz_tree = graphviz.Digraph(comment="policy_tree")
        self.root.add_nodes_to_graphviz_tree(graphviz_tree)
        self.root.add_edges_to_graphviz_tree(graphviz_tree)
        return graphviz_tree



class MdpFamilyResult:
    def __init__(self):
        # if None, then family is undediced
        # if False, then all family members are UNSAT
        # otherwise, contains a satisfying policy for all MDPs in the family
        self.policy = None

        self.game_policy = None
        self.hole_selection = None
        self.splitter = None

class SynthesizerPolicyTree(paynt.synthesizer.synthesizer.Synthesizer):

    # if True, tree leaves will be double-checked after synthesis
    double_check_policy_tree_leaves = False
    # if True, unreachable choices will be discarded from the splitting scheduler
    discard_unreachable_choices = False
    
    @property
    def method_name(self):
        return "AR (policy tree)"

    @staticmethod
    def double_check_policy(quotient, family, prop, policy):
        _,mdp = quotient.fix_and_apply_policy_to_family(family, policy)
        if family.size == 1:
            quotient.assert_mdp_is_deterministic(mdp, family)
        DOUBLE_CHECK_PRECISION = 1e-6
        default_precision = Property.model_checking_precision
        Property.set_model_checking_precision(DOUBLE_CHECK_PRECISION)
        policy_result = mdp.model_check_property(prop, alt=True)
        Property.set_model_checking_precision(default_precision)
        if not policy_result.sat:
            logger.warning("policy should be SAT but (most likely due to model checking precision) has value {}".format(policy_result.value))
        return
    
    
    def verify_policy(self, family, prop, policy):
        _,mdp = self.quotient.fix_and_apply_policy_to_family(family, policy)
        policy_result = mdp.model_check_property(prop, alt=True)
        self.stat.iteration(mdp)
        return policy_result.sat

    
    def solve_singleton(self, family, prop):
        result = family.mdp.model_check_property(prop)
        self.stat.iteration(family.mdp)
        if not result.sat:
            return False
        policy = self.quotient.scheduler_to_policy(result.result.scheduler, family.mdp)
    
        # uncomment below to preemptively double-check the policy
        # SynthesizerPolicyTree.double_check_policy(self.quotient, family, prop, policy)
        return policy


    def solve_game_abstraction(self, family, prop, game_solver):
        # construct and solve the game abstraction
        # logger.debug("solving game abstraction...")

        game_solver.solve_sg(family.selected_choices)
        # game_solver.solve_smg(family.selected_choices)

        game_value = game_solver.solution_value
        self.stat.iteration_game(family.mdp.states)
        game_sat = prop.satisfies_threshold_within_precision(game_value)
        # logger.debug("game solved, value is {}".format(game_value))
        game_policy = game_solver.solution_state_to_player1_action
        # fix irrelevant choices
        game_policy_fixed = self.quotient.empty_policy()
        for state,action in enumerate(game_policy):
            if action < self.quotient.num_actions:
                game_policy_fixed[state] = action
        game_policy = game_policy_fixed
        return game_policy,game_sat

    def state_to_choice_to_hole_selection(self, state_to_choice):
        if SynthesizerPolicyTree.discard_unreachable_choices:
            state_to_choice = self.quotient.discard_unreachable_choices(state_to_choice)
        scheduler_choices = self.quotient.state_to_choice_to_choices(state_to_choice)
        hole_selection = self.quotient.coloring.collectHoleOptions(scheduler_choices)
        return scheduler_choices,hole_selection

    def parse_game_scheduler(self, game_solver):
        state_to_choice = game_solver.solution_state_to_quotient_choice.copy()
        scheduler_choices,hole_selection = self.state_to_choice_to_hole_selection(state_to_choice)
        state_values = game_solver.solution_state_values
        return scheduler_choices,hole_selection,state_values

    def verify_family(self, family, game_solver, prop):
        # logger.info("investigating family of size {}".format(family.size))
        self.quotient.build(family)
        mdp_family_result = MdpFamilyResult()

        if family.size == 1:
            mdp_family_result.policy = self.solve_singleton(family,prop)
            return mdp_family_result
        
        if family.candidate_policy is None:
            game_policy,game_sat = self.solve_game_abstraction(family,prop,game_solver)
        else:
            game_policy = family.candidate_policy
            game_sat = False

        mdp_family_result.game_policy = game_policy
        if game_sat:
            mdp_family_result.policy = game_policy
            return mdp_family_result

        # solve primary direction for the MDP abstraction
        mdp_result = family.mdp.model_check_property(prop)
        mdp_value = mdp_result.value
        self.stat.iteration(family.mdp)
        # logger.debug("primary-primary direction solved, value is {}".format(mdp_value))
        if not mdp_result.sat:
            mdp_family_result.policy = False
            return mdp_family_result

        # undecided: choose scheduler choices to be used for splitting
        scheduler_choices,hole_selection,state_values = self.parse_game_scheduler(game_solver)

        splitter = self.choose_splitter(family,prop,scheduler_choices,state_values,hole_selection)
        mdp_family_result.splitter = splitter
        mdp_family_result.hole_selection = hole_selection
        return mdp_family_result
    
    def choose_splitter(self, family, prop, scheduler_choices, state_values, hole_selection):
        inconsistent_assignments = {hole:options for hole,options in enumerate(hole_selection) if len(options) > 1}
        if len(inconsistent_assignments)==0:
            # # pick any hole with multiple options involved in the hole selection
            for hole,options in enumerate(hole_selection):
                if family.hole_num_options(hole) > 1 and len(options) > 0:
                    return hole
            # pick any hole with multiple options
            # logger.debug("picking an arbitrary hole...")
            for hole in range(family.num_holes):
                if family.hole_num_options(hole) > 1:
                    return hole
        if len(inconsistent_assignments)==1:
            for hole in inconsistent_assignments.keys():
                return hole
        
        # compute scores for inconsistent holes
        scores = self.compute_scores(prop, scheduler_choices, state_values, inconsistent_assignments)
        splitters = self.quotient.holes_with_max_score(scores)
        splitter = splitters[0]
        return splitter
    
    def compute_scores(self, prop, scheduler_choices, state_values, inconsistent_assignments):
        mdp = self.quotient.quotient_mdp
        choice_values = self.quotient.choice_values(mdp, prop, state_values)
        expected_visits = None
        if self.quotient.compute_expected_visits:
            expected_visits = self.quotient.compute_expected_visits(mdp, prop, scheduler_choices)
        quotient_choice_map = [choice for choice in range(self.quotient.quotient_mdp.nr_choices)]
        scores = self.quotient.estimate_scheduler_difference(self.quotient.quotient_mdp, quotient_choice_map, inconsistent_assignments, choice_values, expected_visits)
        return scores

    def assign_candidate_policy(self, subfamilies, hole_selection, splitter, policy):
        policy_consistent = all([len(options) <= 1 for options in hole_selection])
        if not policy_consistent:
            return
        # associate the branch of the split that contains hole selection with the policy
        used_options = hole_selection[splitter]
        if len(used_options) != 1:
            # not sure what to do in this case
            return
        option = used_options[0]
        for subfamily in subfamilies:
            if option in subfamily.hole_options(splitter):
                subfamily.candidate_policy = policy
                return

    def split(self, family, prop, hole_selection, splitter, policy):
        # split the hole
        used_options = hole_selection[splitter]
        if len(used_options) > 1:
            # used_options = used_options[0:1]
            core_suboptions = [[option] for option in used_options]
            other_suboptions = [option for option in family.hole_options(splitter) if option not in used_options]
            if other_suboptions:
                other_suboptions = [other_suboptions]
            else:
                other_suboptions = []
            suboptions = other_suboptions + core_suboptions # DFS solves core first
        else:
            options = family.hole_options(splitter)
            assert len(options) > 1
            half = len(options) // 2
            suboptions = [options[:half], options[half:]]

        subfamilies = family.split(splitter,suboptions)
        for subfamily in subfamilies:
            subfamily.candidate_policy = None

        if not SynthesizerPolicyTree.discard_unreachable_choices:
            self.assign_candidate_policy(subfamilies, hole_selection, splitter, policy)

        return suboptions,subfamilies

    
    def evaluate_all(self, family, prop, keep_value_only=False):
        assert not prop.reward, "expecting reachability probability propery"
        game_solver = self.quotient.build_game_abstraction_solver(prop)
        family.candidate_policy = None
        policy_tree = PolicyTree(family)

        undecided_leaves = [policy_tree.root]
        while undecided_leaves:

            # gi = self.stat.iterations_game
            # if gi is not None and gi > 1000:
            #     return None

            policy_tree_node = undecided_leaves.pop(-1)
            family = policy_tree_node.family
            result = self.verify_family(family,game_solver,prop)
            family.candidate_policy = None

            if result.policy is not None:
                self.explore(family)
                if policy_tree_node != policy_tree.root:
                    family.mdp = None
                if result.policy is False:
                    policy_tree_node.sat = False
                else:
                    policy_tree_node.sat = True
                    policy_tree_node.policy_index = policy_tree.new_policy(result.policy)
                continue

            # refine
            suboptions,subfamilies = self.split(family, prop, result.hole_selection, result.splitter, result.game_policy)
            if policy_tree_node != policy_tree.root:
                family.mdp = None
            policy_tree_node.split(result.splitter,suboptions,subfamilies)
            undecided_leaves += policy_tree_node.child_nodes

        if SynthesizerPolicyTree.double_check_policy_tree_leaves:
            policy_tree.double_check(self.quotient, prop)
        policy_tree.print_stats()

        self.stat.num_mdps_total = self.quotient.family.size
        self.stat.num_mdps_sat = sum([n.family.size for n in policy_tree.collect_sat()])
        self.stat.num_nodes = len(policy_tree.collect_all())
        self.stat.num_leaves = len(policy_tree.collect_leaves())
        self.stat.num_policies = len(policy_tree.policies)
        postprocessing_time = policy_tree.postprocess(self.quotient, prop)
        policy_tree.print_stats()
        self.stat.postprocessing_time = postprocessing_time
        self.stat.num_nodes_merged = len(policy_tree.collect_all())
        self.stat.num_leaves_merged = len(policy_tree.collect_leaves())
        self.stat.num_policies_merged = len(policy_tree.policies)
        self.policy_tree = policy_tree

        # convert policy tree to family evaluation
        evaluations = []
        for node in policy_tree.collect_leaves():
            policy = policy_tree.policies[node.policy_index] if node.sat else None
            evaluation = paynt.synthesizer.synthesizer.FamilyEvaluation(node.family,None,node.sat,policy=policy)
            evaluations.append(evaluation)
        return evaluations


    def run(self, optimum_threshold=None):
        return self.evaluate()


    def export_evaluation_result(self, evaluations, export_filename_base):
        import json
        policies = self.policy_tree.extract_policies(self.quotient)
        policies_json = {}
        for index,key_value in enumerate(policies.items()):
            policy_id,policy = key_value
            policy_json = self.quotient.policy_to_json(policy)
            policies_json[policy_id] = policy_json
        policies_string = json.dumps(policies_json, indent=4)

        policies_filename = export_filename_base + ".json"
        with open(policies_filename, 'w') as file:
            file.write(policies_string)

        logger.info(f"exported policies to {policies_filename}")

        tree = self.policy_tree.extract_policy_tree(self.quotient)
        tree_filename = export_filename_base + ".dot"
        with open(tree_filename, 'w') as file:
            file.write(tree.source)
        logger.info(f"exported policy tree to {tree_filename}")

        tree_visualization_filename = export_filename_base + ".png"
        tree.render(export_filename_base, format="png", cleanup=True) # using export_filename_base since graphviz appends .png by default
        logger.info(f"exported policy tree visualization to {tree_visualization_filename}")



import paynt.synthesizer.synthesizer
import paynt.synthesizer.conflict_generator.dtmc
import paynt.synthesizer.conflict_generator.mdp
import paynt.family.smt

import logging
logger = logging.getLogger(__name__)


class SynthesizerCEGIS(paynt.synthesizer.synthesizer.Synthesizer):

    # CLI argument selecting conflict generator
    conflict_generator_type = None

    def __init__(self, quotient):
        super().__init__(quotient)

        self.conflict_generator = self.choose_conflict_generator(quotient)

        # assert that no reward formula is maximizing
        assert not self.quotient.specification.contains_maximizing_reward_properties, \
            "Cannot use CEGIS for maximizing reward formulae -- consider using AR or hybrid methods."

    
    def choose_conflict_generator(self, quotient):
        if SynthesizerCEGIS.conflict_generator_type == "mdp":
            conflict_generator = paynt.synthesizer.conflict_generator.mdp.ConflictGeneratorMdp(quotient)
        else:
            # default conflict generator
            conflict_generator = paynt.synthesizer.conflict_generator.dtmc.ConflictGeneratorDtmc(quotient)
        return conflict_generator

    
    @property
    def method_name(self):
        return "CEGIS " + self.conflict_generator.name

    
    def collect_conflict_requests(self, family, mc_result):
        '''
        Construct conflict request wrt each unsatisfiable property,
            pack such properties as well as their MDP results (if available)
        '''
        conflict_requests = []
        for index in family.constraint_indices:
            member_result = mc_result.constraints_result.results[index]
            if member_result.sat:
                continue
            prop = self.quotient.specification.constraints[index]
            family_result = family.analysis_result.constraints_result.results[index] if family.analysis_result is not None else None
            conflict_requests.append( (index,prop,family_result) )
        if self.quotient.specification.has_optimality:
            member_result = mc_result.optimality_result
            index = len(self.quotient.specification.constraints)
            prop = self.quotient.specification.optimality
            family_result = family.analysis_result.optimality_result if family.analysis_result is not None else None
            conflict_requests.append( (index,prop,family_result) )

        return conflict_requests


    def analyze_family_assignment_cegis(self, family, assignment):
        """
        :return (1) list of conflicts to exclude from design space (might be empty)
        :return (2) accepting assignment (or None)
        """
        assert family.mdp is not None, "analyzed family does not have an associated quotient MDP"

        dtmc = self.quotient.build_assignment(assignment)
        self.stat.iteration(dtmc)
        result = dtmc.check_specification(self.quotient.specification, family.constraint_indices, short_evaluation=True)
        # analyze model checking results
        accepting_assignment = None
        accepting,improving_value = result.accepting_dtmc(self.quotient.specification)
        if accepting:
            accepting_assignment = assignment
        if improving_value is not None:
            self.quotient.specification.optimality.update_optimum(improving_value)
        if accepting and not self.quotient.specification.can_be_improved():
            return [], accepting_assignment

        conflict_requests = self.collect_conflict_requests(family, result)
        conflicts = self.conflict_generator.construct_conflicts(family, assignment, dtmc, conflict_requests)

        return conflicts, accepting_assignment


    def synthesize_one(self, family):

        # build the quotient, map mdp states to hole indices
        self.quotient.build(family)
        self.conflict_generator.initialize()

        # use sketch design space as a SAT baseline (TODO why?)
        smt_solver = paynt.family.smt.SmtSolver(self.quotient.family)
        
        # CEGIS loop
        assignment = smt_solver.pick_assignment(family)
        while assignment is not None:
            
            conflicts, accepting_assignment = self.analyze_family_assignment_cegis(family, assignment)
            if accepting_assignment is not None:
                self.best_assignment = accepting_assignment
                if not self.quotient.specification.can_be_improved():
                    return self.best_assignment

            pruned = smt_solver.exclude_conflicts(family, assignment, conflicts)
            self.explored += pruned
            
            # construct next assignment
            assignment = smt_solver.pick_assignment(family)
        return self.best_assignment



import paynt.models
import paynt.models.models
import payntbind
import stormpy

import paynt.quotient.quotient
import paynt.quotient.pomdp

import paynt.verification.property

import logging
logger = logging.getLogger(__name__)


class PosmgQuotient(paynt.quotient.quotient.Quotient):

    # implicit size for POSMG unfolding
    initial_memory_size = 1

    def __init__(self, posmg, specification):
        super().__init__(specification = specification)


        # defualt POSMG model
        self.posmg = posmg

        # unfolded POSMG
        self.quotient_mdp = None
        self.family = None
        self.coloring = None

        # number of actions available at each optimizing player observation
        self.actions_at_opt_player_observation = None
        # action labels corresponding to ^
        self.action_labels_at_opt_player_observation = None
        # action labels at each posmg state
        self.action_labels_at_posmg_state = None
        # number of observations of optimizing player
        self.opt_player_observation_count = self.posmg.get_p0_observation_count()
        # for each optimizing player observation, number of states associated with it
        self.opt_player_observation_states = None
        # number of memory states allocated to each optimizing player observation
        self.opt_player_observation_memory_size = None
        # POSMG manager
        self.posmg_manager = None

        # for each optimizing player observation, a list of action holes
        self.opt_player_observation_action_holes = None
        # for each optimizing player observation, a list of memory holes
        self.opt_player_observation_memory_holes = None
        # for each hole, an indication whether this is an action or a memory hole
        self.is_action_hole = None

        # state player indications
        state_players = self.posmg.get_state_player_indications()
        # all observation
        state_obs = self.posmg.get_observations()

        if not specification.has_optimality:
            self.optimizing_player = 0
        else:
            self.optimizing_player = specification.optimality.game_optimizing_player


        # initialize posmg manager
        self.posmg_manager = payntbind.synthesis.PosmgManager(self.posmg, self.optimizing_player)

        # optimizing player observations
        self.opt_player_observations = self.posmg_manager.get_observation_mapping()


        # compute actions available at each optimizing player observation
        self.actions_at_opt_player_observation = {obs:0 for obs in self.opt_player_observations}
        for state in range(self.posmg.nr_states):
            if state_players[state] == self.optimizing_player:
                obs = state_obs[state]
                if self.actions_at_opt_player_observation[obs] != 0:
                    assert self.posmg.get_nr_available_actions(state) == self.actions_at_opt_player_observation[obs]
                    continue
                self.actions_at_opt_player_observation[obs] = self.posmg.get_nr_available_actions(state)

        # collect lables of action available for each prototype
        self.action_labels_at_posmg_state = [[] for state in range(self.posmg.nr_states)]
        for state in range(self.posmg.nr_states):
            if self.action_labels_at_posmg_state[state] != []:
                continue
            actions = self.posmg.get_nr_available_actions(state)
            for offset in range(actions):
                choice = self.posmg.get_choice_index(state, offset)
                labels = self.posmg.choice_labeling.get_labels_of_choice(choice)
                assert len(labels) <= 1, "expected at most 1 label"
                if len(labels) == 0:
                    label = self.EMPTY_LABEL
                else:
                    label = list(labels)[0]
                self.action_labels_at_posmg_state[state].append(label)

        # collect labels of actions available at each observation
        self.action_labels_at_opt_player_observation = {}
        for state in range(self.posmg.nr_states):
            if state_players[state] == self.optimizing_player:
                obs = state_obs[state]
                labels = self.action_labels_at_posmg_state[state]
                self.action_labels_at_opt_player_observation[obs] = labels

        # mark perfect observations
        #self.opt_player_observation_states = [0 for _ in range(self.opt_player_observation_count)]
        self.opt_player_observation_states = {obs:0 for obs in self.opt_player_observations}
        for state in range(self.posmg.nr_states):
            if state_players[state] == self.optimizing_player:
                obs = state_obs[state]
                self.opt_player_observation_states[obs] += 1

        self.set_imperfect_memory_size(PosmgQuotient.initial_memory_size)
        self.current_memory_size = PosmgQuotient.initial_memory_size



    def set_manager_memory_vector(self):
        for obs, memory in self.opt_player_observation_memory_size.items():
            self.posmg_manager.set_observation_memory_size(obs, memory)

    def set_imperfect_memory_size(self, memory_size):
        self.opt_player_observation_memory_size = { obs:(memory_size if obs_states > 1 else 1) for obs, obs_states in self.opt_player_observation_states.items() }

        self.set_manager_memory_vector()
        self.unfold_memory()



    def create_hole_name(self, player, value, mem, is_action_hole):
        category = "A" if is_action_hole else "M"
        # obs_label = self.observation_labels[obs] # TODO maybe we will have this in the future
        if player == self.optimizing_player:
            return "{}(P{},O{},M{})".format(category,player,value,mem)
        else:
            return "{}(P{},S{},M{})".format(category,player,value,mem)

    # version where each state of non optimising players has it's own action hole
    def create_coloring(self):

        # create holes
        family = paynt.family.family.Family()
        # self.observation_action_holes = [] # TODO maybe needed?
        # self.observation_memory_holes = [] # TODO maybe needed?
        # self.is_action_hole = [] # TODO maybe needed?

        for opt_player_obs in self.opt_player_observations:

            # action holes
            hole_indices = []
            num_actions = self.actions_at_opt_player_observation[opt_player_obs]
            if num_actions > 1:
                option_labels = self.action_labels_at_opt_player_observation[opt_player_obs]
                for mem in range(self.opt_player_observation_memory_size[opt_player_obs]):
                    hole_indices.append(family.num_holes)
                    name = self.create_hole_name(self.optimizing_player,opt_player_obs,mem,True)
                    family.add_hole(name,option_labels)
                    # self.is_action_hole.append(True)
            # self.observation_action_holes.append(hole_indices)

            # memory holes
            hole_indices = []
            num_updates = self.posmg_manager.max_successor_memory_size[opt_player_obs]
            if num_updates > 1:
                option_labels = [str(x) for x in range(num_updates)]
                for mem in range(self.opt_player_observation_memory_size[opt_player_obs]):
                    name = self.create_hole_name(self.optimizing_player,opt_player_obs,mem,False)
                    hole_indices.append(family.num_holes)
                    family.add_hole(name,option_labels)
                    # self.is_action_hole.append(False)
            # self.observation_memory_holes.append(hole_indices)

        for state in range(self.quotient_mdp.nr_states):
            quotient_state_player_indication = self.posmg_manager.get_state_player_indications()
            if quotient_state_player_indication[state] != self.optimizing_player:
                num_actions = self.posmg_manager.get_action_count(state)
                if num_actions > 1:
                    posmg_state = self.posmg_manager.state_prototype[state]
                    mem = self.posmg_manager.state_memory[state]
                    option_labels = self.action_labels_at_posmg_state[posmg_state]
                    hole_indices.append(family.num_holes)
                    name = self.create_hole_name(quotient_state_player_indication[state],posmg_state,mem,True)
                    family.add_hole(name,option_labels)

        # create the coloring
        assert self.posmg_manager.num_holes == family.num_holes
        num_holes = family.num_holes
        choice_action_hole = self.posmg_manager.row_action_hole
        choice_memory_hole = self.posmg_manager.row_memory_hole
        choice_action_option = self.posmg_manager.row_action_option
        choice_memory_option = self.posmg_manager.row_memory_option
        choice_to_hole_options = []
        for choice in range(self.quotient_mdp.nr_choices):
            hole_options = []
            hole = choice_action_hole[choice]
            if hole != num_holes:
                hole_options.append( (hole,choice_action_option[choice]) )
            hole = choice_memory_hole[choice]
            if hole != num_holes:
                hole_options.append( (hole,choice_memory_option[choice]) )
            choice_to_hole_options.append(hole_options)

        return family, choice_to_hole_options


    def unfold_memory(self):

        # reset attributes
        self.quotient_mdp = None
        self.coloring = None
        self.hole_option_to_actions = None

        # self.observation_action_holes = None
        # self.observation_memory_holes = None
        # self.is_action_hole = None

        logger.debug("unfolding {}-FSC template into one-sided POSMG...".format(max(self.opt_player_observation_memory_size.values())))
        self.quotient_mdp = self.posmg_manager.construct_mdp()
        self.choice_destinations = payntbind.synthesis.computeChoiceDestinations(self.quotient_mdp)
        logger.debug(f"constructed quotient MDP having {self.quotient_mdp.nr_states} states and {self.quotient_mdp.nr_choices} actions.")

        self.family, choice_to_hole_options = self.create_coloring()

        self.coloring = payntbind.synthesis.Coloring(self.family.family, self.quotient_mdp.nondeterministic_choice_indices, choice_to_hole_options)

        # to each hole-option pair a list of actions colored by this combination
        self.hole_option_to_actions = [[] for hole in range(self.family.num_holes)]
        for hole in range(self.family.num_holes):
            self.hole_option_to_actions[hole] = [[] for option in self.family.hole_options(hole)]
        for choice in range(self.quotient_mdp.nr_choices):
            for hole,option in choice_to_hole_options[choice]:
                self.hole_option_to_actions[hole][option].append(choice)


    def create_smg_from_mdp(self, mdp):
        quotient_player_indications = self.posmg_manager.get_state_player_indications()

        transition_matrix = mdp.model.transition_matrix
        state_labeling = mdp.model.labeling
        components = stormpy.SparseModelComponents(
            transition_matrix=transition_matrix,
            state_labeling=state_labeling)

        if mdp.model.has_choice_labeling():
            components.choice_labeling = mdp.model.choice_labeling

        state_player_inidcations = []
        for state in range(mdp.states):
            quotient_state = mdp.quotient_state_map[state]
            player = quotient_player_indications[quotient_state]
            state_player_inidcations.append(player)
        components.state_player_indications = state_player_inidcations

        return paynt.models.models.Smg(stormpy.storage.SparseSmg(components))

    def scheduler_selection(self, mdp, scheduler):
        ''' Get hole options involved in the scheduler selection. '''
        assert scheduler.memoryless and scheduler.deterministic
        state_to_choice = self.scheduler_to_state_to_choice(mdp, scheduler, discard_unreachable_choices=False)
        choices = self.state_to_choice_to_choices(state_to_choice)
        hole_selection = self.coloring.collectHoleOptions(choices)
        return hole_selection



import stormpy
import payntbind

import paynt.family.family
import paynt.models.models

import math
import itertools

import logging
logger = logging.getLogger(__name__)


class Quotient:

    # if True, expected visits will not be computed for hole scoring
    disable_expected_visits = False

    # label associated with un-labelled choices
    EMPTY_LABEL = "__no_label__"

    @staticmethod
    def make_vector_defined(vector):
        vector_noinf = [ value if value != math.inf else 0 for value in vector]
        default_value = sum(vector_noinf) / len(vector)
        vector_valid = [ value if value != math.inf else default_value for value in vector]
        return vector_valid

    def __init__(self, quotient_mdp = None, family = None, coloring = None, specification = None):

        # colored qoutient MDP for the super-family
        self.quotient_mdp = quotient_mdp
        self.family = family
        self.coloring = coloring
        self.specification = specification

        # builder options
        self.subsystem_builder_options = stormpy.SubsystemBuilderOptions()
        self.subsystem_builder_options.build_state_mapping = True
        self.subsystem_builder_options.build_action_mapping = True

        # for each choice of the quotient, a list of its state-destinations
        self.choice_destinations = None
        if self.quotient_mdp is not None:
            self.choice_destinations = payntbind.synthesis.computeChoiceDestinations(self.quotient_mdp)


    def export_result(self, dtmc):
        ''' to be overridden '''
        pass


    def restrict_mdp(self, mdp, choices):
        '''
        Restrict the quotient MDP to the selected actions.
        :param choices a bitvector of selected actions
        :return (1) the restricted model
        :return (2) sub- to full state mapping
        :return (3) sub- to full action mapping
        '''
        keep_unreachable_states = False # TODO investigate this
        all_states = stormpy.BitVector(mdp.nr_states, True)
        submodel_construction = stormpy.construct_submodel(
            mdp, all_states, choices, keep_unreachable_states, self.subsystem_builder_options
        )
        model = submodel_construction.model
        state_map = submodel_construction.new_to_old_state_mapping.copy()
        choice_map = submodel_construction.new_to_old_action_mapping.copy()
        return model,state_map,choice_map

    def restrict_quotient(self, choices):
        return self.restrict_mdp(self.quotient_mdp, choices)

    def build_from_choice_mask(self, choices):
        mdp,state_map,choice_map = self.restrict_quotient(choices)
        return paynt.models.models.SubMdp(mdp, state_map, choice_map)

    def build(self, family):
        ''' Construct the quotient MDP for the family. '''
        # select actions compatible with the family and restrict the quotient
        choices = self.coloring.selectCompatibleChoices(family.family)
        family.mdp = self.build_from_choice_mask(choices)
        family.selected_choices = choices
        family.mdp.family = family


    @staticmethod
    def mdp_to_dtmc(mdp):
        tm = mdp.transition_matrix
        tm.make_row_grouping_trivial()
        assert tm.nr_columns == tm.nr_rows, "expected transition matrix without non-trivial row groups"
        if mdp.is_exact:
            components = stormpy.storage.SparseExactModelComponents(tm, mdp.labeling, mdp.reward_models)
            dtmc = stormpy.storage.SparseExactDtmc(components)
            return dtmc
        else:
            components = stormpy.storage.SparseModelComponents(tm, mdp.labeling, mdp.reward_models)
            dtmc = stormpy.storage.SparseDtmc(components)
            return dtmc

    def build_assignment(self, family):
        assert family.size == 1, "expecting family of size 1"
        choices = self.coloring.selectCompatibleChoices(family.family)
        assert choices.number_of_set_bits() > 0
        mdp,state_map,choice_map = self.restrict_quotient(choices)
        model = Quotient.mdp_to_dtmc(mdp)
        return paynt.models.models.SubMdp(model,state_map,choice_map)

    def empty_scheduler(self):
        return [None] * self.quotient_mdp.nr_states

    def discard_unreachable_choices(self, state_to_choice):
        state_to_choice_reachable = self.empty_scheduler()
        state_visited = [False]*self.quotient_mdp.nr_states
        initial_state = list(self.quotient_mdp.initial_states)[0]
        state_visited[initial_state] = True
        state_queue = [initial_state]
        while state_queue:
            state = state_queue.pop()
            choice = state_to_choice[state]
            state_to_choice_reachable[state] = choice
            for dst in self.choice_destinations[choice]:
                if not state_visited[dst]:
                    state_visited[dst] = True
                    state_queue.append(dst)
        return state_to_choice_reachable

    def scheduler_to_state_to_choice(self, submdp, scheduler, discard_unreachable_choices=True):
        if submdp.model.is_exact:
            state_to_quotient_choice = payntbind.synthesis.schedulerToStateToGlobalChoiceExact(scheduler, submdp.model, submdp.quotient_choice_map)
        else:
            state_to_quotient_choice = payntbind.synthesis.schedulerToStateToGlobalChoice(scheduler, submdp.model, submdp.quotient_choice_map)
        state_to_choice = self.empty_scheduler()
        for state in range(submdp.model.nr_states):
            quotient_choice = state_to_quotient_choice[state]
            quotient_state = submdp.quotient_state_map[state]
            state_to_choice[quotient_state] = quotient_choice
        if discard_unreachable_choices:
            state_to_choice = self.discard_unreachable_choices(state_to_choice)
        return state_to_choice

    def state_to_choice_to_choices(self, state_to_choice):
        num_choices = self.quotient_mdp.nr_choices
        choices = stormpy.BitVector(num_choices,False)
        for choice in state_to_choice:
            if choice is not None and choice < num_choices:
                choices.set(choice,True)
        return choices

    def scheduler_selection(self, mdp, scheduler):
        ''' Get hole options involved in the scheduler selection. '''
        assert scheduler.memoryless and scheduler.deterministic
        state_to_choice = self.scheduler_to_state_to_choice(mdp, scheduler)
        choices = self.state_to_choice_to_choices(state_to_choice)
        hole_selection = self.coloring.collectHoleOptions(choices)
        return hole_selection


    def choice_values(self, mdp, prop, state_values):
        '''
        Get choice values after model checking MDP against a property.
        Value of choice c: s -> s' is computed as
        rew(c) + sum_s' [ P(s,c,s') * mc(s') ], where
        - rew(c) is the reward associated with choice (c)
        - P(s,c,s') is the probability of transitioning from s to s' under action c
        - mc(s') is the model checking result in state s'
        '''

        # multiply probability with model checking results
        if mdp.is_exact:
            choice_values = payntbind.synthesis.multiply_with_vector_exact(mdp.transition_matrix, state_values)
        else:
            choice_values = payntbind.synthesis.multiply_with_vector(mdp.transition_matrix, state_values)
        choice_values = Quotient.make_vector_defined(choice_values)

        # if the associated reward model has state-action rewards, then these must be added to choice values
        if prop.reward:
            reward_name = prop.formula.reward_name
            rm = mdp.reward_models.get(reward_name)
            assert rm.has_state_action_rewards
            choice_rewards = list(rm.state_action_rewards)
            assert mdp.nr_choices == len(choice_rewards)
            for choice in range(mdp.nr_choices):
                choice_values[choice] += choice_rewards[choice]

        return choice_values


    def compute_expected_visits(self, mdp, prop, choices):
        '''
        Compute expected number of visits in the states of DTMC induced by the shoices.
        '''
        if Quotient.disable_expected_visits:
            return [1]*self.quotient_mdp.nr_states

        # extract DTMC induced by this MDP-scheduler
        sub_mdp,state_map,_ = self.restrict_mdp(mdp, choices)
        dtmc = Quotient.mdp_to_dtmc(sub_mdp)
        dtmc_visits = paynt.verification.property.Property.compute_expected_visits(dtmc)

        # handle infinity- and zero-visits
        if prop.minimizing:
            dtmc_visits = Quotient.make_vector_defined(dtmc_visits)
        else:
            dtmc_visits = [ value if value != math.inf else 0 for value in dtmc_visits]

        # map vector of expected visits onto the state space of the quotient MDP
        expected_visits = [0] * mdp.nr_states
        for state in range(dtmc.nr_states):
            mdp_state = state_map[state]
            visits = dtmc_visits[state]
            expected_visits[mdp_state] = visits

        return expected_visits


    def estimate_scheduler_difference(self, mdp, quotient_choice_map, inconsistent_assignments, choice_values, expected_visits):
        hole_variance = payntbind.synthesis.computeInconsistentHoleVariance(
            self.family.family, mdp.nondeterministic_choice_indices, quotient_choice_map, choice_values,
            self.coloring, inconsistent_assignments, expected_visits)
        return hole_variance


    def scheduler_is_consistent(self, mdp, prop, result):
        '''
        Get hole assignment induced by this scheduler and fill undefined
        holes by some option from the design space of this mdp.
        :return hole assignment
        :return whether the scheduler is consistent
        '''
        if mdp.is_deterministic:
            selection = [[mdp.family.hole_options(hole)[0]] for hole in range(mdp.family.num_holes)]
            return selection, True

        # get qualitative scheduler selection, filter inconsistent assignments
        selection = self.scheduler_selection(mdp, result.scheduler)
        inconsistent_assignments = {hole:options for hole,options in enumerate(selection) if len(options) > 1 }
        scheduler_is_consistent = len(inconsistent_assignments) == 0
        for hole,options in enumerate(selection):
            if len(options) == 0:
                # if some hole options are not involved in the selection, we can fix an arbitrary value
                selection[hole] = [mdp.family.hole_options(hole)[0]]

        return selection, scheduler_is_consistent

    def scheduler_scores(self, mdp, prop, result, selection):
        inconsistent_assignments = {hole:options for hole,options in enumerate(selection) if len(options) > 1 }
        choice_values = self.choice_values(mdp.model, prop, result.get_values())
        choices = result.scheduler.compute_action_support(mdp.model.nondeterministic_choice_indices)
        expected_visits = self.compute_expected_visits(mdp.model, prop, choices)
        scores = self.estimate_scheduler_difference(mdp.model, mdp.quotient_choice_map, inconsistent_assignments, choice_values, expected_visits)
        return scores


    def suboptions_half(self, mdp, splitter):
        ''' Split options of a splitter into two halves. '''
        options = mdp.family.hole_options(splitter)
        half = len(options) // 2
        suboptions = [options[:half], options[half:]]
        return suboptions

    def suboptions_unique(self, mdp, splitter, used_options):
        ''' Distribute used options of a splitter into different suboptions. '''
        assert len(used_options) > 1
        suboptions = [[option] for option in used_options]
        index = 0
        for option in mdp.family.hole_options(splitter):
            if option in used_options:
                continue
            suboptions[index].append(option)
            index = (index + 1) % len(suboptions)
        return suboptions

    def suboptions_enumerate(self, mdp, splitter, used_options):
        assert len(used_options) > 1
        core_suboptions = [[option] for option in used_options]
        other_suboptions = [option for option in mdp.family.hole_options(splitter) if option not in used_options]
        return core_suboptions, other_suboptions

    def holes_with_max_score(self, hole_score):
        max_score = max(hole_score.values())
        with_max_score = [hole_index for hole_index in hole_score if hole_score[hole_index] == max_score]
        return with_max_score

    def split(self, family):

        mdp = family.mdp
        assert not mdp.is_deterministic

        # split family wrt last undecided result
        result = family.analysis_result.undecided_result()
        hole_assignments = result.primary_selection
        scores = self.scheduler_scores(mdp, result.prop, result.primary.result, result.primary_selection)
        if scores is None:
            scores = {hole:0 for hole in range(mdp.family.num_holes) if mdp.family.hole_num_options(hole) > 1}

        splitters = self.holes_with_max_score(scores)
        splitter = splitters[0]
        if len(hole_assignments[splitter]) > 1:
            core_suboptions,other_suboptions = self.suboptions_enumerate(mdp, splitter, hole_assignments[splitter])
        else:
            assert mdp.family.hole_num_options(splitter) > 1
            core_suboptions = self.suboptions_half(mdp, splitter)
            other_suboptions = []
        # print(mdp.family[splitter], core_suboptions, other_suboptions)

        if len(other_suboptions) == 0:
            suboptions = core_suboptions
        else:
            suboptions = [other_suboptions] + core_suboptions  # DFS solves core first

        # construct corresponding subfamilies
        parent_info = family.collect_parent_info(self.specification)
        subfamilies = family.split(splitter,suboptions)
        for subfamily in subfamilies:
            subfamily.add_parent_info(parent_info)
        return subfamilies


    def get_property(self):
        assert self.specification.num_properties == 1, "expecting a single property"
        return self.specification.all_properties()[0]

    @classmethod
    def identify_absorbing_states(cls, model):
        state_is_absorbing = [True] * model.nr_states
        tm = model.transition_matrix
        for state in range(model.nr_states):
            for choice in tm.get_rows_for_group(state):
                for entry in tm.get_row(choice):
                    if entry.column != state:
                        state_is_absorbing[state] = False
                        break
                if not state_is_absorbing[state]:
                    break
        return state_is_absorbing

    @classmethod
    def identify_states_with_actions(cls, model):
        ''' Get a mask of states having more than one action. '''
        state_has_actions = [None] * model.nr_states
        ndi = model.nondeterministic_choice_indices
        for state in range(model.nr_states):
            num_actions = ndi[state+1]-ndi[state]
            state_has_actions[state] = (num_actions>1)
        return state_has_actions

    def identify_target_states(self, model=None, prop=None):
        if model is None:
            model = self.quotient_mdp
        if prop is None:
            prop = self.get_property()
        if prop.is_discounted_reward:
            return stormpy.BitVector(model.nr_states,False)
        target_label = prop.get_target_label()
        return model.labeling.get_states(target_label)



import stormpy

import paynt.verification.property
import paynt.verification.property_result

import payntbind

import logging
logger = logging.getLogger(__name__)


class Mdp:

    @classmethod
    def assert_no_overlapping_guards(cls, model):
        if model.labeling.contains_label("overlap_guards"):
            assert model.labeling.get_states("overlap_guards").number_of_set_bits() == 0

    def __init__(self, model):
        # Mdp.assert_no_overlapping_guards(model)
        self.model = model
        if len(model.initial_states) > 1:
            logger.warning("WARNING: obtained model with multiple initial states")

    @property
    def states(self):
        return self.model.nr_states

    @property
    def is_deterministic(self):
        return self.model.nr_choices == self.model.nr_states

    @property
    def initial_state(self):
        return self.model.initial_states[0]

    def model_check_property(self, prop, alt=False):
        formula = prop.formula if not alt else prop.formula_alt
        result = paynt.verification.property.Property.model_check(self.model,formula)
        value = result.at(self.initial_state)
        return paynt.verification.property_result.PropertyResult(prop, result, value)

    def check_specification(self, spec, constraint_indices=None, short_evaluation=False):
        ''' Assuming this is a DTMC. '''
        if constraint_indices is None:
            constraint_indices = spec.all_constraint_indices()
        results = [None for _ in spec.constraints]
        for index in constraint_indices:
            result = self.model_check_property(spec.constraints[index])
            results[index] = result
            if short_evaluation and result.sat is False:
                break
        spec_result = paynt.verification.property_result.SpecificationResult()
        spec_result.constraints_result = paynt.verification.property_result.ConstraintsResult(results)

        if spec.has_optimality and not (short_evaluation and spec_result.constraints_result.sat is False):
            spec_result.optimality_result = self.model_check_property(spec.optimality)
        return spec_result



class SubMdp(Mdp):

    def __init__(self, model, quotient_state_map, quotient_choice_map):
        super().__init__(model)
        self.quotient_choice_map = quotient_choice_map
        self.quotient_state_map = quotient_state_map


class Smg(Mdp):

    def __init__(self, model):
        super().__init__(model)

    def model_check_property(self, prop, alt=False):
        formula = prop.game_formula if not alt else prop.game_formula_alt

        result = payntbind.synthesis.model_check_smg(self.model, formula,
                                                        only_initial_states=False, set_produce_schedulers=True,
                                                        env=paynt.verification.property.Property.environment)

        value = result.at(self.model.initial_states[0])
        return paynt.verification.property_result.PropertyResult(prop, result, value)



import stormpy
import payntbind

import paynt
import paynt.quotient.quotient

import logging
logger = logging.getLogger(__name__)

class DecPomdpQuotient(paynt.quotient.quotient.Quotient):

    # implicit initial FSC size
    initial_memory_size = 1

    def __init__(self, decpomdp_manager, specification):
        super().__init__(specification = specification)

        assert decpomdp_manager.num_agents > 1

        self.decpomdp_manager = decpomdp_manager

        # for each agent a (simplified) label for each observation
        self.agent_observation_labels = decpomdp_manager.agent_observation_labels
        # for each agent its action labels
        self.agent_action_labels = decpomdp_manager.agent_action_labels
        # number of agent in the model
        self.nr_agents = decpomdp_manager.num_agents
        # for each joint observation contains observation of each agent
        self.joint_observations = decpomdp_manager.joint_observations
        # for each state its corresponding joint observation index
        self.state_joint_observation = decpomdp_manager.state_joint_observation
        # for each agent the number of its observations
        self.nr_agent_observations = [len(observation) for observation in self.agent_observation_labels]
        # number of states in the dec-pomdp model
        self.nr_states = decpomdp_manager.num_decpomdp_states()
        # for each agent number of states with the given observation
        self.agent_observation_states = None
        # for each agent and each observation the size of the memory, this is used to set the memory in the quotient
        self.agent_observation_memory_size = [[] for _ in range(self.nr_agents)]
        # for each agent the number of available actions at given observation
        self.num_agent_actions_at_observation = decpomdp_manager.num_agent_actions_at_observation

        # for each hole, an indication whether this is an action or a memory hole
        self.is_action_hole = None

        # mark perfectly observable states for each agent
        self.agent_observation_states = [[0 for obs in range(self.nr_agent_observations[agent])] for agent in range(self.nr_agents)]
        for state in range(self.nr_states):
            joint_observation = self.state_joint_observation[state]
            for agent in range(self.nr_agents):
                agent_obs = self.joint_observations[joint_observation][agent]
                self.agent_observation_states[agent][agent_obs] += 1
                
        # do initial unfolding
        self.set_imperfect_memory_size(DecPomdpQuotient.initial_memory_size)
        self.current_memory_size = DecPomdpQuotient.initial_memory_size
        

    def create_hole_name(self, agent, obs, mem, is_action_hole):
        category = "A" if is_action_hole else "M"
        obs_label = self.agent_observation_labels[agent][obs]
        return "{}({},{},{})".format(category,agent,obs_label,mem)

    def set_imperfect_memory_size(self, memory_size):
        ''' Set given memory size only to imperfect observations. '''
        for agent in range(self.nr_agents):
            agent_memory = [memory_size if self.agent_observation_states[agent][obs] > 1 else 1 for obs in range(self.nr_agent_observations[agent])]
            self.agent_observation_memory_size[agent] = agent_memory

        self.set_manager_memory_vector()
        self.unfold_memory()

    def set_agent_imperfect_memory_size(self, agent, memory_size):
        ''' Set given memory size only to imperfect observations of given agent. '''
        assert agent in range(self.nr_agents), "given agent index is larger than number of agents"

        agent_memory = [memory_size if self.agent_observation_states[agent][obs] > 1 else 1 for obs in range(self.nr_agent_observations[agent])]
        self.agent_observation_memory_size[agent] = agent_memory

        self.set_manager_memory_vector()
        self.unfold_memory()

    def set_manager_memory_vector(self):
        for agent, agent_memory in enumerate(self.agent_observation_memory_size):
            for obs, memory in enumerate(agent_memory):
                self.decpomdp_manager.set_agent_observation_memory_size(agent, obs, memory)
            

    def substitute_suffix(self, string, delimiter, replacer):
        '''Subsitute the suffix behind the last delimiter.'''
        output_string = string.split(delimiter)
        output_string[-1] = str(replacer)
        output_string = delimiter.join(output_string)
        return output_string

    def unfold_memory(self):
        
        # reset attributes
        self.quotient_mdp = None
        self.coloring = None
        self.hole_option_to_actions = None
        
        self.is_action_hole = None
        
        self.quotient_mdp = self.decpomdp_manager.construct_quotient_mdp()
        self.choice_destinations = payntbind.synthesis.computeChoiceDestinations(self.quotient_mdp)
        logger.debug(f"constructed quotient MDP having {self.quotient_mdp.nr_states} states and {self.quotient_mdp.nr_choices} actions.")

        self.family, choice_to_hole_options = self.create_coloring()
       
        self.coloring = payntbind.synthesis.Coloring(self.family.family, self.quotient_mdp.nondeterministic_choice_indices, choice_to_hole_options)

        # to each hole-option pair a list of actions colored by this combination
        self.hole_option_to_actions = [[] for hole in range(self.family.num_holes)]
        for hole in range(self.family.num_holes):
            self.hole_option_to_actions[hole] = [[] for option in self.family.hole_options(hole)]
        for choice in range(self.quotient_mdp.nr_choices):
            for hole,option in choice_to_hole_options[choice]:
                self.hole_option_to_actions[hole][option].append(choice)


    def create_coloring(self):
        # short aliases
        pm = self.decpomdp_manager

        # create holes
        family = paynt.family.family.Family()
        self.is_action_hole = []

        for agent in range(self.nr_agents):
            for obs in range(self.nr_agent_observations[agent]):
                
                # action holes
                hole_indices = []
                num_actions = self.num_agent_actions_at_observation[agent][obs]
                if num_actions > 1:
                    option_labels = ["act_"+str(x) for x in range(num_actions)] # TODO adding labels here would be nice in the future
                    for mem in range(self.agent_observation_memory_size[agent][obs]):
                        hole_indices.append(family.num_holes)
                        name = self.create_hole_name(agent,obs,mem,True)
                        family.add_hole(name,option_labels)
                        self.is_action_hole.append(True)

                # memory holes
                hole_indices = []
                num_updates = pm.agent_max_successor_memory_size[agent][obs]
                if num_updates > 1:
                    option_labels = [str(x) for x in range(num_updates)]
                    for mem in range(self.agent_observation_memory_size[agent][obs]):
                        name = self.create_hole_name(agent,obs,mem,False)
                        hole_indices.append(family.num_holes)
                        family.add_hole(name,option_labels)
                        self.is_action_hole.append(False)

        # create the coloring
        assert pm.num_holes == family.num_holes
        num_holes = family.num_holes
        agent_row_action_hole = pm.agent_row_action_hole
        agent_row_action_option = pm.agent_row_action_option
        agent_row_memory_hole = pm.agent_row_memory_hole
        agent_row_memory_option = pm.agent_row_memory_option
        choice_to_hole_options = []
        for choice in range(self.quotient_mdp.nr_choices):
            hole_options = []
            for agent in range(self.nr_agents): 
                hole = agent_row_action_hole[agent][choice]
                if hole != num_holes:
                    hole_options.append( (hole,agent_row_action_option[agent][choice]) )
                hole = agent_row_memory_hole[agent][choice]
                if hole != num_holes:
                    hole_options.append( (hole,agent_row_memory_option[agent][choice]) )
            choice_to_hole_options.append(hole_options)
        return family, choice_to_hole_options

    def estimate_scheduler_difference(self, mdp, quotient_choice_map, inconsistent_assignments, choice_values, expected_visits=None):
        return super().estimate_scheduler_difference(mdp, quotient_choice_map, inconsistent_assignments, choice_values, expected_visits)


import stormpy
import payntbind

import paynt.models.model_builder
import paynt.quotient.quotient
import paynt.quotient.mdp
import paynt.quotient.pomdp
import paynt.quotient.decpomdp
import paynt.quotient.posmg
import paynt.quotient.mdp_family
import paynt.quotient.pomdp_family
import paynt.verification.property

from paynt.parser.prism_parser import PrismParser
from paynt.parser.drn_parser import DrnParser

import uuid
import os
import json

import logging
logger = logging.getLogger(__name__)


def substitute_suffix(string, delimiter, replacer):
    '''Subsitute the suffix behind the last delimiter.'''
    output_string = string.split(delimiter)
    output_string[-1] = str(replacer)
    output_string = delimiter.join(output_string)
    return output_string


def make_rewards_action_based(model):
    tm = model.transition_matrix
    for name,reward_model in model.reward_models.items():
        assert not reward_model.has_transition_rewards, "Paynt does not support transition rewards"
        if not reward_model.has_state_rewards:
            continue
        logger.info("converting state rewards '{}' to state-action rewards".format(name))
        if reward_model.has_state_action_rewards:
            logger.info("state rewards will be added to existing state-action rewards".format(name))
            action_reward = reward_model.state_action_rewards.copy()
        else:
            action_reward = [0] * model.nr_choices

        for state in range(model.nr_states):
            state_reward = reward_model.get_state_reward(state)
            for action in range(tm.get_row_group_start(state),tm.get_row_group_end(state)):
                action_reward[action] += state_reward

        if model.is_exact:
            payntbind.synthesis.remove_reward_model_exact(model,name)
            new_reward_model = stormpy.storage.SparseExactRewardModel(optional_state_action_reward_vector=action_reward)
            model.add_reward_model(name, new_reward_model)
        else:
            payntbind.synthesis.remove_reward_model(model,name)
            new_reward_model = stormpy.storage.SparseRewardModel(optional_state_action_reward_vector=action_reward)
            model.add_reward_model(name, new_reward_model)

class Sketch:

    @classmethod
    def load_sketch(cls, sketch_path, properties_path,
        export=None, relative_error=0, precision=1e-4, constraint_bound=None, use_exact=False):

        prism = None
        explicit_quotient = None
        specification = None
        family = None
        coloring = None
        jani_unfolder = None
        decpomdp_manager = None
        obs_evaluator = None

        paynt.verification.property.Property.model_checking_precision = precision

        # check path
        if not os.path.isfile(sketch_path):
            raise ValueError(f"the sketch file {sketch_path} does not exist")
        logger.info(f"loading sketch from {sketch_path} ...")

        filetype = None
        try:
            logger.info(f"assuming sketch in PRISM format...")
            prism, explicit_quotient, specification, family, coloring, jani_unfolder, obs_evaluator = PrismParser.read_prism(
                        sketch_path, properties_path, relative_error, use_exact)
            filetype = "prism"
        except SyntaxError:
            pass
        if filetype is None:
            try:
                logger.info(f"assuming sketch in DRN format...")
                explicit_quotient = paynt.models.model_builder.ModelBuilder.from_drn(sketch_path, use_exact)
                specification = PrismParser.parse_specification(properties_path, relative_error, use_exact=use_exact)
                filetype = "drn"
                project_path = os.path.dirname(sketch_path)
                valuations_filename = "state-valuations.json"
                valuations_path = project_path + "/" + valuations_filename
                state_valuations = None
                if os.path.exists(valuations_path) and os.path.isfile(valuations_path):
                    with open(valuations_path) as file:
                        state_valuations = json.load(file)
                if state_valuations is not None:
                    if use_exact:
                        raise Exception("exact synthesis is not supported with state valuations")
                    logger.info(f"found state valuations in {valuations_path}, adding to the model...")
                    explicit_quotient = payntbind.synthesis.addStateValuations(explicit_quotient,state_valuations)
            except Exception as e:
                print(e)
                pass
        if filetype is None:
            try:
                logger.info(f"assuming sketch in Cassandra format...")
                decpomdp_manager = payntbind.synthesis.parse_decpomdp(sketch_path)
                if constraint_bound is not None:
                    decpomdp_manager.set_constraint(constraint_bound)
                if decpomdp_manager is None:
                    raise SyntaxError
                logger.info("applying discount factor transformation...")
                decpomdp_manager.apply_discount_factor_transformation()
                explicit_quotient = decpomdp_manager.construct_pomdp()
                if constraint_bound is not None:
                    specification = PrismParser.parse_specification(properties_path, relative_error)
                else:
                    optimality = paynt.verification.property.construct_reward_property(
                        decpomdp_manager.reward_model_name,
                        decpomdp_manager.reward_minimizing,
                        decpomdp_manager.discount_sink_label)
                    specification = paynt.verification.property.Specification([optimality])
                filetype = "cassandra"
            except SyntaxError:
                pass

        assert filetype is not None, "unknown format of input file"
        logger.info("sketch parsing OK")

        paynt.verification.property.Property.initialize(use_exact)
        if explicit_quotient.is_exact:
            updated = payntbind.synthesis.addMissingChoiceLabelsExact(explicit_quotient)
        else:
            updated = payntbind.synthesis.addMissingChoiceLabels(explicit_quotient)
        if updated is not None: explicit_quotient = updated
        if not payntbind.synthesis.assertChoiceLabelingIsCanonic(explicit_quotient.nondeterministic_choice_indices,explicit_quotient.choice_labeling,False):
            logger.warning("WARNING: choice labeling for the quotient is not canonic")


        make_rewards_action_based(explicit_quotient)
        logger.debug("constructed explicit quotient having {} states and {} choices".format(
            explicit_quotient.nr_states, explicit_quotient.nr_choices))

        if specification.contains_until_properties() and filetype != "prism":
            logger.info("WARNING: using until formulae with non-PRISM inputs might lead to unexpected behaviour")
        specification.transform_until_to_eventually()
        logger.info(f"found the following specification {specification}")

        if export is not None:
            Sketch.export(export, sketch_path, jani_unfolder, explicit_quotient)
            logger.info("export OK, aborting...")
            exit(0)

        if jani_unfolder is not None:
            if prism.model_type == stormpy.storage.PrismModelType.DTMC:
                quotient_container = paynt.quotient.quotient.Quotient(explicit_quotient, family, coloring, specification)
            elif prism.model_type == stormpy.storage.PrismModelType.MDP:
                quotient_container = paynt.quotient.mdp_family.MdpFamilyQuotient(explicit_quotient, family, coloring, specification)
            elif prism.model_type == stormpy.storage.PrismModelType.POMDP:
                quotient_container = paynt.quotient.pomdp_family.PomdpFamilyQuotient(explicit_quotient, family, coloring, specification, obs_evaluator)
        else:
            # assert explicit_quotient.is_nondeterministic_model, "expected nondeterministic model"
            if decpomdp_manager is not None and decpomdp_manager.num_agents > 1:
                quotient_container = paynt.quotient.decpomdp.DecPomdpQuotient(decpomdp_manager, specification)
            elif isinstance(explicit_quotient, payntbind.synthesis.Posmg):
                quotient_container = paynt.quotient.posmg.PosmgQuotient(explicit_quotient, specification)
            elif not explicit_quotient.is_partially_observable:
                quotient_container = paynt.quotient.mdp.MdpQuotient(explicit_quotient, specification)
            else:
                quotient_container = paynt.quotient.pomdp.PomdpQuotient(explicit_quotient, specification, decpomdp_manager)
        return quotient_container


    @classmethod
    def load_sketch_as_all_in_one(cls, sketch_path, properties_path):
        if not os.path.isfile(sketch_path):
            raise ValueError(f"the sketch file {sketch_path} does not exist")
        logger.info(f"loading sketch from {sketch_path} ...")
        logger.info(f"all in one approach so assuming input in PRISM format...")
        try:
            prism, hole_definitions = PrismParser.load_sketch_prism(sketch_path)
            expression_parser = stormpy.storage.ExpressionParser(prism.expression_manager)
            expression_parser.set_identifier_mapping(dict())
            prism_model_type = {
                stormpy.storage.PrismModelType.DTMC:"DTMC",
                stormpy.storage.PrismModelType.MDP:"MDP",
                stormpy.storage.PrismModelType.POMDP:"POMDP"
            }[prism.model_type]
            logger.debug("PRISM model type: " + prism_model_type)

            hole_expressions = None
            family = None
            if len(hole_definitions) > 0:
                logger.info("processing hole definitions...")
                prism, hole_expressions, family = PrismParser.parse_holes(prism, expression_parser, hole_definitions)

            specification = PrismParser.parse_specification(properties_path, prism=prism)

            prism = prism.replace_variable_initialization_by_init_expression()
            expression_manager = prism.expression_manager
            for index, hole in enumerate(hole_definitions):
                # TODO add support for double holes
                assert hole[1] == 'int', "all in one approach only works with integer holes"
                var = prism.get_constant(hole[0])
                var_values = [x.evaluate_as_int() for x in hole_expressions[index]]
                prism = prism.replace_constant_by_variable(var, expression_manager.create_integer(min(var_values)), expression_manager.create_integer(max(var_values)))
                var_options = [stormpy.Expression.Eq(var.expression_variable.get_expression(), expression_manager.create_integer(val)) for val in var_values]
                prism.update_initial_states_expression(stormpy.Expression.And(prism.initial_states_expression, stormpy.Expression.Disjunction(var_options)))

            # TODO investigate why we have to print and load the prism program again for all in one construction to work
            tmp_path = sketch_path + str(uuid.uuid4())
            with open(tmp_path, 'w') as f:
                print(prism, end="", file=f)
            try:
                prism = stormpy.parse_prism_program(tmp_path, prism_compat=True)
                os.remove(tmp_path)
            except:
                os.remove(tmp_path)
                raise SyntaxError

        except SyntaxError as e:
            logger.error(f"all in one approach supports only input in PRISM format!")
            raise e

        return prism, specification, family

    @classmethod
    def export(cls, export, sketch_path, jani_unfolder, explicit_quotient):
        if export == "jani":
            assert jani_unfolder is not None, "jani unfolder was not used"
            output_path = substitute_suffix(sketch_path, '.', 'jani')
            jani_unfolder.write_jani(output_path)
        if export == "drn":
            output_path = substitute_suffix(sketch_path, '.', 'drn')
            stormpy.export_to_drn(explicit_quotient, output_path)
        if export == "pomdp":
            assert explicit_quotient.is_nondeterministic_model and explicit_quotient.is_partially_observable, \
                "cannot '--export pomdp' with non-POMDP sketches"
            output_path = substitute_suffix(sketch_path, '.', 'pomdp')
            property_path = substitute_suffix(sketch_path, '/', 'props.pomdp')
            paynt.parser.pomdp_parser.PomdpParser.write_model_in_pomdp_solve_format(explicit_quotient, output_path, property_path)



import stormpy
import payntbind

from helpers.helper import get_stormpy_example_path, get_builder_options


class TestModelTranslation:

    def test_add_dont_care_action(self):
        # setup
        program = stormpy.parse_prism_program(get_stormpy_example_path("mdp", "tiny_rewards.nm"), prism_compat=True)
        program = program.label_unlabelled_commands({}) # this is something we always do in PAYNT but without it this will throw segfault!
        mdp = stormpy.build_sparse_model_with_options(program, get_builder_options())
        state_bitvector = stormpy.BitVector(mdp.nr_states, True)

        # test
        mdp_with_dont_care_action = payntbind.synthesis.addDontCareAction(mdp, state_bitvector)

        # assert
        assert mdp_with_dont_care_action.choice_labeling.contains_label("__random__")
        assert mdp_with_dont_care_action.nr_states == mdp.nr_states
        assert mdp_with_dont_care_action.nr_choices == mdp.nr_choices + mdp.nr_states

        nci = mdp_with_dont_care_action.nondeterministic_choice_indices.copy()
        reward_model = mdp_with_dont_care_action.reward_models['']
        for state in range(mdp_with_dont_care_action.nr_states):
            probs = [0.0,0.0,0.0]
            rews = 0.0
            num_non_dont_care_choices = nci[state + 1] - nci[state] - 1
            for choice in range(nci[state], nci[state + 1]):
                if "__random__" in mdp_with_dont_care_action.choice_labeling.get_labels_of_choice(choice):
                    for entry in mdp_with_dont_care_action.transition_matrix.get_row(choice):
                        state = entry.column
                        prob = entry.value()
                        assert prob == probs[state] / num_non_dont_care_choices
                        assert reward_model.state_action_rewards[choice] == rews / num_non_dont_care_choices 
                else:
                    for entry in mdp_with_dont_care_action.transition_matrix.get_row(choice):
                        state = entry.column
                        prob = entry.value()
                        probs[state] += prob
                        rews += reward_model.state_action_rewards[choice]

    def test_add_dont_care_action_exact(self):
        # setup
        program = stormpy.parse_prism_program(get_stormpy_example_path("mdp", "tiny_rewards.nm"), prism_compat=True)
        program = program.label_unlabelled_commands({}) # this is something we always do in PAYNT but without it this will throw segfault!
        mdp = stormpy.build_sparse_exact_model_with_options(program, get_builder_options())
        state_bitvector = stormpy.BitVector(mdp.nr_states, True)

        # test
        mdp_with_dont_care_action = payntbind.synthesis.addDontCareActionExact(mdp, state_bitvector)

        # assert
        assert mdp_with_dont_care_action.choice_labeling.contains_label("__random__")
        assert mdp_with_dont_care_action.nr_states == mdp.nr_states
        assert mdp_with_dont_care_action.nr_choices == mdp.nr_choices + mdp.nr_states

        nci = mdp_with_dont_care_action.nondeterministic_choice_indices.copy()
        reward_model = mdp_with_dont_care_action.reward_models['']
        for state in range(mdp_with_dont_care_action.nr_states):
            probs = [0,0,0]
            rews = 0
            num_non_dont_care_choices = nci[state + 1] - nci[state] - 1
            for choice in range(nci[state], nci[state + 1]):
                if "__random__" in mdp_with_dont_care_action.choice_labeling.get_labels_of_choice(choice):
                    for entry in mdp_with_dont_care_action.transition_matrix.get_row(choice):
                        state = entry.column
                        prob = entry.value()
                        assert prob == probs[state] / num_non_dont_care_choices
                        assert reward_model.state_action_rewards[choice] == rews / num_non_dont_care_choices 
                else:
                    for entry in mdp_with_dont_care_action.transition_matrix.get_row(choice):
                        state = entry.column
                        prob = entry.value()
                        probs[state] += prob
                        rews += reward_model.state_action_rewards[choice]

    # TODO add more tests
    # TODO also add test for exact models




import stormpy

from paynt.verification.property import *


class PropertyResult:
    def __init__(self, prop, result, value):
        self.result = result
        self.value = value
        self.sat = prop.satisfies_threshold(value)
        self.improves_optimum = None if not isinstance(prop,OptimalityProperty) else prop.improves_optimum(value)

    def __str__(self):
        return str(self.value)


class ConstraintsResult:
    '''
    A list of constraint results.
    Note: some results might be None (not evaluated).
    '''
    def __init__(self, results):
        self.results = results
        self.undecided_constraints = [i for i,result in enumerate(results) if result is not None and result.sat is None]
        
        result_sats = [result.sat for result in results if result is not None]
        if False in result_sats:
            self.sat = False
        elif None in result_sats:
            self.sat = None
        else:
            self.sat = True


    def __str__(self):
        return ",".join([str(result) for result in self.results])


class SpecificationResult:
    def __init__(self):
        self.constraints_result = None
        self.optimality_result = None

    def __str__(self):
        return str(self.constraints_result) + " : " + str(self.optimality_result)

    def accepting_dtmc(self, specification):
        """
        :return (1) whether specification is satisfied (dtmc is accepting)
        :return (2) new optimal value associated with the accepting dtmc
            (can be None if no optimality)
        """

        if not self.constraints_result.sat:
            # constraints not sat
            return False, None

        if self.optimality_result is None:
            # constraints sat and no optimality
            return True, None
        
        if not self.optimality_result.improves_optimum:
            # constraints sat and optimality not sat
            return False, None
        
        # constraints sat and optimality sat
        return True, self.optimality_result.value


    def undecided_result(self):
        if self.optimality_result is not None and self.optimality_result.can_improve:
            return self.optimality_result
        return self.constraints_result.results[self.constraints_result.undecided_constraints[0]]

            

class MdpPropertyResult:
    def __init__(self,prop):
        self.prop = prop
        self.primary = None
        self.secondary = None
        self.sat = None
        self.primary_selection = None

    @property
    def minimizing(self):
        return self.prop.minimizing

    def __str__(self):
        prim = str(self.primary)
        seco = str(self.secondary)
        if self.minimizing:
            return "{} - {}".format(prim,seco)
        else:
            return "{} - {}".format(seco,prim)



class MdpOptimalityResult(MdpPropertyResult):
    def __init__(self, prop):
        super().__init__(prop)
        self.improving_assignment = None
        self.improving_value = None
        self.can_improve = None


class MdpSpecificationResult(SpecificationResult):

    def __init__(self):
        super().__init__()

    def evaluate(self, family=None, admissible_assignment=None):
        self.improving_assignment = None
        self.improving_value = None
        self.can_improve = None

        cr = self.constraints_result
        opt = self.optimality_result

        if cr.sat is False:
            self.can_improve = False
            return

        if cr.sat is True:
            # all constraints were satisfied
            if opt is None:
                if admissible_assignment is not None:
                    self.improving_assignment = admissible_assignment
                else:
                    self.improving_assignment = family
                self.can_improve = False
            else:
                self.improving_assignment = opt.improving_assignment
                self.improving_value = opt.improving_value
                self.can_improve = opt.can_improve
            return

        # constraints undecided
        if opt is None: 
            self.can_improve = True
        else:
            self.improving_assignment = opt.improving_assignment
            self.improving_value = opt.improving_value
            self.can_improve = opt.improving_value is None and opt.can_improve



from paynt.synthesizer.synthesizer import Synthesizer
from paynt.synthesizer.synthesizer_ar import SynthesizerAR

import os
import time
import multiprocessing

import logging
logger = logging.getLogger(__name__)


import cProfile, pstats

# global variables
# when a new process is spawned (forked), it will inherit these variables from the parent
quotient = None
profiler = None

# helper functions for family serialization
def family_to_hole_options(family):
    return [family.hole_options(hole) for hole in range(family.num_holes)]

def hole_options_to_family(hole_options):
    family = quotient.family.copy()
    for hole,options in enumerate(hole_options):
        family.hole_set_options(hole,options)
    return family


def solve_family(args):
    '''
    Build the quotient, analyze it and, if necessary, split into subfamilies.
    '''
    try:

        if args is None:
            pstats.Stats(profiler).sort_stats('tottime').print_stats(10)
            return

        hole_options, optimum = args
        # re-construct the family
        family = hole_options_to_family(hole_options)

        # synchronize optimum
        if optimum is not None:
            quotient.specification.optimality.optimum = optimum

        quotient.build(family)
        quotient.check_specification_for_mdp(family)
        res = family.analysis_result
        improving_value = res.improving_value
        improving_assignment = res.improving_assignment
        if improving_assignment is not None:
            improving_assignment = family_to_hole_options(improving_assignment)

        subfamilies = []
        if res.can_improve:
            subfamilies = quotient.split(family)
            subfamilies = [ family_to_hole_options(family) for family in subfamilies ]

        return (family.mdp.states, improving_value, improving_assignment, subfamilies)

    except:
        logger.error("Worker sub-process encountered an error.")
        return None



class SynthesizerMultiCoreAR(SynthesizerAR):

    @property
    def method_name(self):
        return "AR (multicore)"

    def synthesize_one(self, family):

        satisfying_assignment = None
        families = [family]

        global quotient
        quotient = self.quotient
        profiling = False
        if profiling:
            global profiler
            profiler = cProfile.Profile()
            profiler.enable()

        # create a pool of processes
        # by default, os.cpu_count() processes will be spawned
        with multiprocessing.Pool(
            # processes=1
        ) as pool:

            while families:

                # get current optimum
                optimum = None
                if self.quotient.specification.has_optimality:
                    optimum = self.quotient.specification.optimality.optimum

                # work with some number of families
                # print("submitting ", len(families), " families")

                split = os.cpu_count() * 1
                input_families = families[-split:]
                input_families_size = sum([family.size for family in input_families])
                remaining_families = families[:-split]
                input_families = [family_to_hole_options(family) for family in input_families]

                inputs = zip(input_families, [optimum] * len(input_families))

                results = pool.map(solve_family, inputs)
                # results = pool.imap_unordered(solve_family, inputs, chunksize=10)
                # results = pool.map(solve_batch, inputs)

                # process the results
                new_families = []
                for r in results:
                    if r is None:
                        logger.error("Worker sub-process encountered an error.")
                        exit()
                    mdp_states, improving_value, improving_assignment, subfamilies_hole_options = r
                    self.stat.iteration_mdp(mdp_states)

                    if improving_value is not None:
                        if self.quotient.specification.optimality.improves_optimum(improving_value):
                            self.quotient.specification.optimality.update_optimum(improving_value)
                            improving_assignment = hole_options_to_family(improving_assignment)
                            satisfying_assignment = improving_assignment

                    subfamilies = [hole_options_to_family(hole_options) for hole_options in subfamilies_hole_options]
                    new_families += subfamilies

                new_families_size = sum([family.size for family in new_families])
                self.explored += input_families_size-new_families_size

                families = remaining_families + new_families

            if profiling:
                pool.apply(solve_family, (None,))

        if profiling:
            pstats.Stats(profiler).sort_stats('tottime').print_stats(10)
        return satisfying_assignment



import payntbind.info
import stormpy.info

import logging
logger = logging.getLogger(__name__)

# Checks whether stormpy and payntbind use the same Storm as backend
# Note that even if all these checks pass there is still a chance that the version of Storm used are incompatible but this should be rare
def check_stormpy_compatibility():

    incompatibility_found = False

    if payntbind.info.storm_version() != stormpy.info.storm_version():
        logger.warning(f"Storm used by payntbind ({payntbind.info.storm_version()}) and stormpy ({stormpy.info.storm_version()}) are not the same!")
        incompatibility_found = True
    if payntbind.info.storm_from_system() != stormpy.info.storm_from_system():
        logger.warning(f"Storm used by payntbind ({'system' if payntbind.info.storm_from_system() else 'fetched'}) and stormpy ({'system' if stormpy.info.storm_from_system() else 'fetched'}) have different origins!")
        incompatibility_found = True
    if payntbind.info.storm_development_version() != stormpy.info.storm_development_version():
        logger.warning(f"Storm used by payntbind ({'development' if payntbind.info.storm_development_version() else 'stable'}) and stormpy ({'development' if stormpy.info.storm_development_version() else 'stable'}) have different development versions!")
        incompatibility_found = True
    if payntbind.info.storm_origin_info() != stormpy.info.storm_origin_info():
        logger.warning(f"Storm used by payntbind ({payntbind.info.storm_origin_info()}) and stormpy ({stormpy.info.storm_origin_info()}) have different origin information!")
        incompatibility_found = True
    if payntbind.info.storm_directory() != stormpy.info.storm_directory():
        logger.warning(f"Storm used by payntbind is located at {payntbind.info.storm_directory()} while Storm used by stormpy is located at {stormpy.info.storm_directory()}.")
        incompatibility_found = True

    if not incompatibility_found:
        logger.info(f"Storm used by payntbind and stormpy seem to be the same.")
